""" Manage a data-transformation recipe
Started April 2019 by David Megginson
License: Public Domain
"""

import flask, hxl_proxy, hxl_proxy.dao, hxl_proxy.filters, logging, werkzeug


class Recipe:
    """ Class to hold a HXL Proxy recipe.
    The recipe can come either from the request parameters, or from a saved recipe
    in the database. For a saved recipe, it's still possible to override
    certain properties (especially the URL) with the request parameters, so that
    you can use the same recipe with multiple source URLs (disabled for private
    datasets with authentication tokens).
    """

    RECIPE_OVERRIDES = ['url', 'schema_url', 'stub']
    """ Properties that may be overridden in a saved recipe """


    def __init__(self, recipe_id=None, auth=False, request_args=None):
        """ Recipe constructor
        @param recipe_id: the hash identifier of an existing saved recipe
        @param auth: if true, the user needs to authenticate to use the recipe
        @param request_args: custom args to substitute for the current Flask request
        """

        # initialise the properties
        self.recipe_id = str(recipe_id) if recipe_id is not None else None
        self.args = None
        self.name = None
        self.description = None
        self.cloneable = True
        self.passhash = None
        self.stub = None

        self.overridden = False
        self.auth = auth

        # default to the request GET parameters
        if request_args is None:
            request_args = flask.request.args

        # do we have a saved recipe? if so, then populate from the saved data
        if recipe_id is not None:

            # read the recipe from the database
            saved_recipe = hxl_proxy.dao.recipes.read(self.recipe_id)

            if not saved_recipe:
                raise werkzeug.exceptions.NotFound("No saved recipe for {}".format(recipe_id))

            # populate the class from the saved recipe dict
            self.fromDict(saved_recipe)

            # check if this page requires authorisation
            if auth and not self.check_auth():
                raise werkzeug.exceptions.Unauthorized("Wrong or missing password.")

            # allow overrides *only* if we're not using a private dataset
            # (not sending an HTTP Authorization: header)
            if "authorization_token" not in self.args:
                for key in self.RECIPE_OVERRIDES:
                    if key in request_args:
                        self.overridden = True
                        self.args[key] = request_args[key]

        # we don't have a saved recipe: use the HTTP GET parameters
        else:
            self.args = request_args
            self.stub = request_args.get("stub")


    @property
    def url(self):
        return self.args.get("url")

    
    @property
    def schema_url(self):
        return self.args.get("schema_url")

    
    def fromDict(self, props):
        """ Deserialise this object from a dict """
        self.recipe_id = props.get("recipe_id")
        self.name = props.get("name")
        self.description = props.get("description")
        self.cloneable = props.get("cloneable")
        self.passhash = props.get("passhash")
        self.stub = props.get("stub")
        self.date_created = props.get('date_created')
        self.date_modified = props.get('date_modified')
        self.args = dict(props.get("args"))

        
    def toDict(self):
        """ Serialise this object to a dict """
        return {
            "recipe_id": self.recipe_id,
            "name": self.name,
            "description": self.description,
            "cloneable": self.cloneable,
            "passhash": self.passhash,
            "stub": self.stub,
            "args": self.args,
        }


    def logs(self, level="WARNING"):
        handler = Recipe.LogHandler(level)
        logging.getLogger('hxl').addHandler(handler)
        logging.getLogger('hxl_proxy').addHandler(handler)
        source = hxl_proxy.filters.setup_filters(self)
        try:
            for row in source:
                pass
        except:
            pass
        return handler.messages
            

    def check_auth(self, password=None):
        """ Check whether a users is authorised to access this page.
        @param password: a cleartext password
        @returns: True if the user is authorised.
        """

        # does this recipe require a password?
        if self.passhash:
            # do we have a clear-text password?
            if password:
                session_passhash = hxl_proxy.util.make_md5(password)
                flask.session['passhash'] = session_passhash
            # no password, so look in the session token
            else:
                session_passhash = flask.session.get('passhash')

            # do the password hashes match?
            if self.passhash == session_passhash:
                return True
            else:
                flask.session['passhash'] = None
                flask.flash("Wrong password")
                return False

        # no password required, so always OK
        else:
            return True


    class LogHandler(logging.Handler):

        def __init__(self, level):
            super().__init__(level)
            self.messages = []

        def handle(self, record):
            self.messages.append(record)
        





""" Context managers for caching """

import hxl_proxy, logging, os, redis, requests_cache

logger = logging.getLogger(__name__)
""" Python logger for this module """


class input:
    """Context manager for input caching.

    Uses the app.config REQUEST_CACHE_* variables, including
    REQUEST_CACHE_TYPE, REQUEST_CACHE_NAME, and
    REQUEST_CACHE_TIMEOUT_SECONDS, as defaults.

    Usage:
        with caching.input():
            data = hxl.data(source_url)

        with caching.input(namespace="foo", timeout=3600):
            data = hxl.data(source_url)

    """

    def __init__ (self, namespace=None, timeout=None):
        """ Temporarily enable input caching.

        Args:
            namespace: the cache namespace to use (defaults to app.config["REQUEST_CACHE_NAME"])
            timeout: the duration for caching items, in seconds (defaults to app.config["REQUEST_CACHE_TIMEOUT_SECONDS"])
        """

        config = hxl_proxy.app.config

        if namespace is None:
            self.namespace = config.get('REQUEST_CACHE_NAME', "hxl-proxy-in")
        else:
            self.namespace = namespace
            
        if timeout is None:
            self.timeout = config.get('REQUEST_CACHE_TIMEOUT', 3600)
        else:
            self.timeout = timeout

        self.backend = config.get('REQUEST_CACHE_BACKEND', "memory")

        self.backend_args = config.get('REQUEST_CACHE_EXTRAS', {})

    def __enter__ (self):
        requests_cache.install_cache(self.namespace, backend=self.backend, expire_after=self.timeout, **self.backend_args)

    def __exit__ (self, type, value, traceback):
        requests_cache.uninstall_cache()



""" HTTP controllers for the HXL Proxy

All of the Flask controllers are in this module.

See unit tests in tests/test_controllers.py

Started January 2015 by David Megginson
License: Public Domain
"""

import hxl_proxy
from hxl.input import HXLIOException

from hxl_proxy import admin, app, auth, cache, caching, dao, exceptions, filters, pcodes, preview, recipes, util, validate

import datetime, flask, hxl, io, json, logging, requests, requests_cache, werkzeug, csv, urllib

from hxl.util import logup

logger = logging.getLogger(__name__)
""" Python logger for this module """

from structlog import contextvars, get_logger, wrap_logger
input_logger = wrap_logger(logging.getLogger('hxl.REMOTE_ACCESS'))

SHEET_MAX_NO = 20


########################################################################
# Error handlers
#
# These functions handle exceptions that make it to the top level.
#
# The HXL Proxy uses exceptions for special purposes like redirections
# or authorisation as well as errors.
########################################################################

def handle_default_exception(e):
    """ Error handler: display an error page with various HTTP status codes
    This handler applies to any exception that doesn't have a more-specific
    handler below.
    @param e: the exception being handled
    """

    key_error = {'error_type': type(e).__name__}
    # log a warning for the error/exception that we're handling
    if hasattr(e, 'message'):
        key_error['message'] = e.message
        logger.warning('%s: %s', type(e).__name__, e.message)
    else:
        logger.warning(type(e).__name__)
    logup('Error', key_error, "warning")


    if isinstance(e, requests.exceptions.HTTPError) or isinstance(e, hxl.input.HXLHTMLException):
        status = 404
        e = exceptions.RemoteDataException(e)
    elif isinstance(e, IOError) or isinstance(e, OSError):
        # probably tried to open an inappropriate URL
        status = 403
    elif isinstance(e, werkzeug.exceptions.HTTPException):
        status = e.code
    else:
        status = 500

    # Check the global output_format variable to see if it's HTML or JSON/CSV
    # Use a JSON error format if not HTML
    if flask.g.output_format != 'html':
        response = flask.Response(util.make_json_error(e, status), mimetype='application/json', status=status)
        # add CORS header
        response.headers['Access-Control-Allow-Origin'] = '*'
        return response
    else:
        # Generic HTML error page
        return flask.render_template('error.html', e=e, category=type(e)), status

# Register the general error handler UNLESS we're in debug mode
if not app.config.get('DEBUG'):
    app.register_error_handler(Exception, handle_default_exception)


def handle_redirect_exception(e):
    """ Error handler: catch a redirection exception
    Different parts of the code throw exceptions.RedirectException
    when they need the HXL Proxy to jump to a different page. This is especially
    important for workflow (e.g. if there's no URL, jump to /data/source)
    @param e: the exception being handled
    """
    if e.message:
        flask.flash(e.message)
    return flask.redirect(e.target_url, e.http_code)

# Register the redirect exception handler
app.register_error_handler(exceptions.RedirectException, handle_redirect_exception)


def handle_source_authorization_exception(e):
    """ Error handler: the data source requires authorisation
    This will be triggered when opening a private HDX dataset before
    the user has supplied their authorisation token.
    @param e: the exception being handled
    """
    if e.message:
        flask.flash(e.message)

    # we're using flask.g.recipe_id to handle the case where a saved recipe
    # points to a formerly-public dataset that has suddenly become private
    # normally, it will be None (because there's no saved recipe yet)
    recipe = recipes.Recipe(recipe_id=flask.g.recipe_id)

    # add an extra parameter for the /data/save form to indicate that we
    # want the user to provide an authorisation token
    extras = {
        'need_token': 'on'
    }

    # note whether the resource looked like it came from HDX
    if e.is_ckan:
        extras['is_ckan'] = 'on'

    # redirect to the /data/save page to ask the user for a token
    return flask.redirect(util.data_url_for('data_save', recipe=recipe, extras=extras), 302)

# register the source authorisation handler
app.register_error_handler(hxl.input.HXLAuthorizationException, handle_source_authorization_exception)


def handle_password_required_exception(e):
    """ Error handler: the HXL Proxy saved recipe requires a password login
    Note that this handler triggers on a recipe basis, not a user basis; each
    each saved recipe can potentially have a different password.
    @param e: the exception being handled
    """
    flask.flash("Login required")
    if flask.g.recipe_id:
        destination = flask.request.path
        args = dict(flask.request.args)
        if args:
            destination += "?" + urllib.parse.urlencode(args)
        return flask.redirect(util.data_url_for('data_login', recipe_id=flask.g.recipe_id, extras={"from": destination}), 303)
    else:
        raise Exception("Internal error: login but no saved recipe")

# register the password required handler
app.register_error_handler(werkzeug.exceptions.Unauthorized, handle_password_required_exception)


def handle_ssl_certificate_error(e):
    """ Error handler: SSL certificate error
    Give the user an option to disable SSL certificate verification by redirecting to the /data/source page.
    @param e: the exception being handled
    """
    if flask.g.output_format == "html":
        flask.flash("SSL error. If you understand the risks, you can check \"Don't verify SSL certificates\" to continue.")
        return flask.redirect(util.data_url_for('data_source', recipe=recipes.Recipe()), 302)
    else:
        response = flask.Response(util.make_json_error(e, 400), mimetype='application/json', status=400)
        # add CORS header
        response.headers['Access-Control-Allow-Origin'] = '*'
        return response

# register the SSL certificate verification handler
app.register_error_handler(requests.exceptions.SSLError, handle_ssl_certificate_error)



########################################################################
# Global pre-/post-controller functions
########################################################################

@app.before_request
def before_request():
    """Code to run immediately before the request"""

    # grab the secret key
    app.secret_key = app.config['SECRET_KEY']

    # choose the parameter storage class before parsing the GET parameters
    flask.request.parameter_storage_class = werkzeug.datastructures.ImmutableOrderedMultiDict

    # grab the member error for Humanitarian.ID (not currently used)
    flask.g.member = flask.session.get('member_info')

    # select the default output format (controllers may change it)
    flask.g.output_format='html'



########################################################################
# Top-level page controllers
########################################################################

# has tests
@app.route("/about.html")
def about():
    """ Flask controller: show the about page
    Includes version information for major packages, so that
    we can tell easily what's deployed.
    """
    # include version information for these packages
    releases = {
        'hxl-proxy': hxl_proxy.__version__,
        'libhxl': hxl.__version__,
        'flask': flask.__version__,
        'requests': requests.__version__
    }

    # draw the web page
    return flask.render_template('about.html', releases=releases)



########################################################################
# /data GET controllers
########################################################################

# has tests
@app.route("/data/<recipe_id>/login")
@util.structlogged
def data_login(recipe_id):
    """ Flask controller: log in to work on a saved recipe
    The user will end up here only if they tried to alter a saved
    recipe. They will have to enter the recipe's password to
    continue.
    @param recipe_id: the hash for a saved recipe (or None if working from the command line)
    """
    flask.g.recipe_id = recipe_id # for error handling
    recipe = recipes.Recipe(recipe_id)
    destination = flask.request.args.get('from')
    if not destination:
        destination = util.data_url_for('data_edit', recipe)
    return flask.render_template('data-login.html', recipe=recipe, destination=destination)

# has tests
@app.route("/data/source")
@app.route("/data/<recipe_id>/source")
@util.structlogged
def data_source(recipe_id=None):
    """ Flask controller: choose a new source URL
    @param recipe_id: the hash for a saved recipe (or None if working from the command line)
    """
    flask.g.recipe_id = recipe_id # for error handling
    recipe = recipes.Recipe(recipe_id, auth=True)
    return flask.render_template('data-source.html', recipe=recipe)

# has tests
@app.route("/data/tagger")
@app.route("/data/<recipe_id>/tagger")
@util.structlogged
def data_tagger(recipe_id=None):
    """ Flask controller: add HXL tags to an untagged dataset
    The template will render differently depending on whether the user has selected the
    last row of text headers yet (&header_row), so this is actually two different workflow
    steps.
    @param recipe_id: the hash for a saved recipe (or None if working from the command line)
    """
    flask.g.recipe_id = recipe_id # for error handling

    # Build the recipe from the GET params and/or the database
    recipe = recipes.Recipe(recipe_id, auth=True)

    # Workflow: if there's no source URL, redirect the user to /data/source
    if not recipe.url:
        logup('No URL supplied for /data/tagger; redirecting to /data/source', level='info')
        logger.info("No URL supplied for /data/tagger; redirecting to /data/source")
        flask.flash('Please choose a data source first.')
        return flask.redirect(util.data_url_for('data_source', recipe), 303)

    # We have to collect the following properties manually, because we don't have a complete
    # HXLated recipe to open yet
    header_row = recipe.args.get('header-row')
    if header_row is not None:
        header_row = int(header_row)

    # Set up a 25-row raw-data preview, using make_input from libhxl-python
    preview = []
    i = 0
    for row in util.hxl_make_input(recipe.url, util.make_input_options(recipe.args)):
        # Stop if we get to 25 rows
        if i >= 25:
            break
        else:
            i = i + 1
        if row:
            preview.append(row)

    if header_row is not None:
        mappings = util.clean_tagger_mappings(preview[header_row-1], recipe)
        mappings += [["", ""]] # room for an extra header in the form
    else:
        mappings = []

    # Draw the web page
    return flask.render_template('data-tagger.html', recipe=recipe, preview=preview, header_row=header_row, mappings=mappings)


# has tests
@app.route("/data/edit")
@app.route("/data/<recipe_id>/edit", methods=['GET', 'POST'])
@util.structlogged
def data_edit(recipe_id=None):
    """Flask controller: create or edit a filter pipeline.
    Output for this page is never cached, but input may be.
    @param recipe_id: the hash for a saved recipe (or None if working from the command line)
    """
    flask.g.recipe_id = recipe_id # for error handling

    # Build the recipe from the GET params and/or the database
    recipe = recipes.Recipe(recipe_id, auth=True)

    # Workflow: if there's no source URL, redirect the user to /data/source
    if not recipe.url:
        flask.flash('Please choose a data source first.')
        logup('No URL supplied for /data/edit; redirecting to /data/source', level="info")
        logger.info("No URL supplied for /data/edit; redirecting to /data/source")
        return flask.redirect(util.data_url_for('data_source', recipe), 303)

    # show only a short preview
    max_rows = recipe.args.get('max-rows')
    max_rows = min(int(max_rows), 25) if max_rows is not None else 25

    # check whether we're stripping headers
    show_headers = (recipe.args.get('strip-headers') != 'on')

    # Special handling: if the user has introduced an error in the filters,
    # catch it so that they have an opportunity to change the filters and try to
    # fix it.
    error = None
    try:
        source = preview.PreviewFilter(filters.setup_filters(recipe), max_rows=max_rows)
        source.columns
    except (
            requests.RequestException,
            exceptions.RedirectException,
            hxl.input.HXLAuthorizationException,
            hxl.input.HXLHTMLException
    ) as e1:
        # always pass through these exceptions
        raise e1
    except Exception as e2:
        # logger.exception(e2)
        error = e2
        source = None

    # Figure out how many filter forms to show
    filter_count = 0
    for n in range(1, filters.MAX_FILTER_COUNT):
        if recipe.args.get('filter%02d' % n):
            filter_count = n
    if filter_count < filters.MAX_FILTER_COUNT:
        filter_count += 1

    # Draw the web page
    return flask.render_template(
        'data-recipe.html',
        recipe=recipe,
        source=source,
        error=error,
        show_headers=show_headers,
        filter_count=filter_count
    )


# has tests
@app.route("/data/save")
@app.route("/data/<recipe_id>/save")
@util.structlogged
def data_save(recipe_id=None):
    """ Flask controller: create or update a saved dataset (with a short URL)
    The user will get redirected here automatically if they attempt to open a private
    dataset on HDX (or anywhere else that requires an "Authorization:" HTTP header.
    The controller creates a form that submits the recipe information to be saved
    in the database and identified with a short hash.
    @param recipe_id: the hash for a saved recipe (or None if working from the command line)
    """
    flask.g.recipe_id = recipe_id # for error handling

    # Build the recipe from the GET params and/or the database
    recipe = recipes.Recipe(recipe_id, auth=True)

    # Workflow: if there's no source URL, redirect the user to /data/source
    if not recipe.url:
        flask.flash('Please choose a data source first.')
        logup('No URL supplied for /data/save; redirecting to /data/source', level="info")
        logger.info("No URL supplied for /data/save; redirecting to /data/source")
        return flask.redirect(util.data_url_for('data_source', recipe), 303)

    # Grab controller-specific properties for the template
    need_token = flask.request.args.get('need_token') # we need an authentication token
    is_ckan = flask.request.args.get('is_ckan') # the source looks like CKAN

    # Draw the web page
    return flask.render_template('data-save.html', recipe=recipe, need_token=need_token, is_ckan=is_ckan)

# has tests
@app.route("/data/validate")
@app.route("/data/validate.<format>")
@app.route("/data/<recipe_id>/validate")
@app.route("/data/<recipe_id>/validate.<format>")
@util.structlogged
def data_validate(recipe_id=None, format='html'):
    """ Flask controller: validate a HXL dataset and show the results
    Output options include a web-based HTML dashboard or JSON.
    Output for this page is never cached, but input may be.
    @param recipe_id: the hash for a saved recipe (or None if working from the command line)
    @param format: the selected output format (json or html)
    """
    # Set global variables
    flask.g.recipe_id = recipe_id # for error handling
    flask.g.output_format = format # requested output format

    # Get the recipe
    recipe = recipes.Recipe(recipe_id)

    # Workflow: if there's no source URL, redirect the user to /data/source
    if not recipe.url:
        flask.flash('Please choose a data source first.')
        logup('No URL supplied for /data/validate; redirecting to /data/source', level="info")
        logger.info("No URL supplied for /data/validate; redirecting to /data/source")
        return flask.redirect(util.data_url_for('data_source', recipe), 303)

    # Set up the HXL validation schema
    schema_source = None
    if recipe.schema_url:
        schema_source = util.hxl_data(recipe.schema_url, util.make_input_options(recipe.args))
        logup('Using HXL validation schema', {"schema": recipe.schema_url}, level="info")
        logger.info("Using HXL validation schema at %s", recipe.schema_url)
    else:
        logup('No HXL validation schema specified; using default schema', level="info")
        logger.info("No HXL validation schema specified; using default schema")

    # Run the validation and get a JSON report from libhxl-python
    error_report = hxl.validate(
        filters.setup_filters(recipe),
        schema_source
    )

    # Render the validation results in JSON
    if format == 'json':
        return flask.Response(
            json.dumps(error_report, indent=4),
            mimetype="application/json"
        )

    # Render the validation results in HTML
    else:
        # issue to highlight (HTML only)
        template_name = 'validate-summary.html'
        selected_issue = None

        # Special GET parameters for controlling validation
        severity_level = flask.request.args.get('severity', 'info')
        detail_hash = flask.request.args.get('details', None)

        # if there's a detail_hash, show just that detail in the report
        if detail_hash:
            logup('Showing validation-report detail', level="info")
            logger.info("Showing validation-report detail")
            template_name = 'validate-issue.html'
            for issue in error_report['issues']:
                if issue['rule_id'] == detail_hash:
                    selected_issue = issue
                    break

        # draw the web page
        return flask.render_template(
            template_name,
            recipe=recipe,
            schema_url=recipe.schema_url,
            error_report=error_report,
            issue=selected_issue,
            severity=severity_level
        )


# has tests
@app.route("/data/advanced")
@app.route("/data/<recipe_id>/advanced")
@util.structlogged
def show_advanced(recipe_id=None):
    """ Flask controller: developer page for entering a JSON recipe directly
    This page isn't linked from the HXML Proxy validation, but it's a convenient
    place to experiment with creating JSON-encoded recipes, as described at
    https://github.com/HXLStandard/hxl-proxy/wiki/JSON-recipes
    """
    recipe = recipes.Recipe(recipe_id)
    return flask.render_template("data-advanced.html", recipe=recipe)


# no tests
@app.route("/data/logs")
@app.route("/data/<recipe_id>/logs")
@util.structlogged
def data_logs(recipe_id=None):
    """ Flask controller: show logs for a recipe
    """
    level = flask.request.args.get('level', 'WARNING').upper()
    recipe = recipes.Recipe(recipe_id)
    return flask.render_template("data-logs.html", recipe=recipe, level=level, in_logger=True)


# has tests
@app.route("/data")
@app.route("/data.<flavour>.<format>")
@app.route("/data.<format>")
@app.route("/data/download/<stub>.<flavour>.<format>")
@app.route("/data/download/<stub>.<format>")
@app.route("/data/<recipe_id>.<flavour>.<format>")
@app.route("/data/<recipe_id>.<format>")
@app.route("/data/<recipe_id>/download/<stub>.<flavour>.<format>")
@app.route("/data/<recipe_id>/download/<stub>.<format>")
@app.route("/data/<recipe_id>") # must come last, or it will steal earlier patterns
@cache.cached(key_prefix=util.make_cache_key, unless=util.skip_cache_p)
@util.structlogged
def data_view(recipe_id=None, format="html", stub=None, flavour=None):
    """ Flask controller: render a transformed dataset
    This is the controller that requests will hit most of the time.
    It renders a transformed dataset as an HTML web page, a JSON
    list of lists, or a JSON list of objects, based on the URL, Note that
    the URL patterns above allow for custom-named download files
    as well as generic downloads, hence the wide variety of patterns.

    This controller MUST come after all the other /data controllers, or
    else Flask will get confused.

    This is a tricky controller to understand, for a few reasons:

    1. It can render output in several different formats
    2. It optionally caches the output
    3. It includes a CORS HTTP header
    4. Most of the work happens inside a nested function, to simplify caching
    5. It may specify a download file name, based on the stub property

    Grab a cup of tea, and work your way through the code slowly. :)

    @param recipe_id: the hash for a saved recipe (or None if working from the command line)
    @param format: the selected output format (json or html)
    @param stub: the root filename for download, if supplied
    @param flavour: the JSON flavour, if supplied (will be "objects")
    """
    flask.g.recipe_id = recipe_id # for error handling

    # Use an internal function to generate the output.
    # That simplifies the control flow, so that we can easily
    # capture the output regardless of chosen format, and
    # cache it as needed. Most of the controller's code
    # is inside this function.
    def get_result ():
        """ Internal output-generation function.
        @returns: a Python generator to produce the input incrementally
        """
        flask.g.output_format = format

        # Set up the data source from the recipe
        recipe = recipes.Recipe(recipe_id, auth=False)

        # Workflow: if there's no source URL, redirect the user to /data/source
        if not recipe.url:
            flask.flash('Please choose a data source first.')
            return flask.redirect(util.data_url_for('data_source', recipe), 303)

        # Use input caching if requested
        if util.skip_cache_p():
            source = filters.setup_filters(recipe)
        else:
            with caching.input():
                source = filters.setup_filters(recipe)

        # Parameters controlling the output
        show_headers = (recipe.args.get('strip-headers') != 'on')
        max_rows = recipe.args.get('max-rows', None)

        # Return a generator based on the format requested

        # Render a web page
        if format == 'html':
            # cap output at 5,000 rows for HTML
            max_rows = min(int(max_rows), 5000) if max_rows is not None else 5000
            return flask.render_template(
                'data-view.html',
                source=preview.PreviewFilter(source, max_rows=max_rows),
                recipe=recipe,
                show_headers=show_headers
            )

        # Data formats from here on ...

        # Limit the number of output rows *only* if requested
        if max_rows is not None:
            source = preview.PreviewFilter(source, max_rows=int(max_rows))

        # Render JSON output (list of lists or list of objects)
        if format == 'json':
            response = flask.Response(
                list(
                    source.gen_json(show_headers=show_headers, use_objects=(flavour=='objects'))
                ),
                mimetype='application/json'
            )

        # Render CSV output
        else:
            response = flask.Response(list(source.gen_csv(show_headers=show_headers)), mimetype='text/csv')

        # Include a CORS header for cross-origin data access
        response.headers['Access-Control-Allow-Origin'] = '*'

        # Set the file download name if &stub is present
        if recipe.stub:
            response.headers['Content-Disposition'] = 'attachment; filename={}.{}'.format(recipe.stub, format)

        # Return the response object
        return response

    # end of internal function

    # Get the result and update the cache manually if we're skipping caching.
    result = get_result()

    # Update the cache even if caching is turned off
    # (this might not be working yet)
    if util.skip_cache_p():
        cache.set(util.make_cache_key(), result)

    # return the response object that will render the output
    return result



#########################################################################
# Primary action POST controllers
# These are URLs that are not bookmarkable.
########################################################################

# needs tests
@app.route("/actions/login", methods=['POST'])
@util.structlogged
def do_data_login():
    """ Flask controller: log the user in for a specific dataset.
    Note that this is NOT a user login; it's a dataset login. That
    means that the user will have to re-login if they start working
    on a dataset that has a different password.

    POST parameters:

    from - the origin URL (return there after login)
    password - the clear-text password
    """

    # Note origin page
    destination = flask.request.form.get('from')
    if not destination:
        destination = util.data_url_for('data_view')

    # Just save the password hash in a cookie, but don't do anything with it
    password = flask.request.form.get('password')
    flask.session['passhash'] = util.make_md5(password)

    # Try opening the original page again, with password hash token in the cookie.
    return flask.redirect(destination, 303)

# needs tests
@app.route("/actions/save-recipe", methods=['POST'])
@util.structlogged
def do_data_save():
    """ Flask controller: create or update a saved recipe
    The saved recipe has all of its parameters in the database, and is
    identified by a short hash. The user needs to supply a password
    to edit it.

    Post parameters:

    recipe_id - the short hash identifying the recipe (blank to create a new one)
    name - the recipe' title (optional)
    description - the recipe's long description (optional)
    cloneable - a flag indicating whether a user may clone the saved recipe (optional; defaults to "on")
    stub - a root filename for downloading (optional)
    password - a clear-text password (optional for a new saved recipe)
    password_repeat - repeated clear-text password (optional for a new saved recipe)

    (Will also include all of the other HXL Proxy recipe arguments as hidden parameters)
    """

    # FIXME - move somewhere else
    RECIPE_ARG_EXCLUDES = [
        'cloneable',
        'description',
        'dest',
        'details'
        'name',
        'passhash',
        'password',
        'password-repeat',
        'recipe_id',
        'severity',
        'stub',
    ]
    """Properties that should never appear in a recipe's args dictionary"""

    # We will have a recipe_id if we're updating an existing pipeline
    recipe_id = flask.request.form.get('recipe_id')
    flask.g.recipe_id = recipe_id # for error handling
    recipe = recipes.Recipe(recipe_id, auth=True, request_args=flask.request.form)

    destination_facet = flask.request.form.get('dest', 'data_view')

    # Update recipe metadata
    # Note that an empty/unchecked value will be omitted from the form
    if 'name' in flask.request.form:
        recipe.name = util.forbid_markup(flask.request.form['name'], 'name')

    if 'description' in flask.request.form:
        recipe.description = util.forbid_markup(flask.request.form['description'], 'description')
    else:
        recipe.description = ''

    if 'cloneable' in flask.request.form and not flask.request.form.get('authorization_token') and flask.request.form['cloneable'] == "on":
        recipe.cloneable = True
    else:
        recipe.cloneable = False

    if 'stub' in flask.request.form:
        recipe.stub = util.forbid_markup(flask.request.form['stub'], 'stub')
    else:
        recipe.stub = ''

    # Merge changed values
    recipe.args = {}
    for name in flask.request.form:
        if name not in RECIPE_ARG_EXCLUDES:
            recipe.args[name] = flask.request.form.get(name)

    # Check for a password change
    password = flask.request.form.get('password')
    password_repeat = flask.request.form.get('password-repeat')

    # Updating an existing recipe.
    if recipe_id:
        if password:
            if password == password_repeat:
                recipe.passhash = util.make_md5(password)
                flask.session['passhash'] = recipe.passhash
            else:
                raise werkzeug.exceptions.BadRequest("Passwords don't match")
        dao.recipes.update(recipe.toDict())

    # Creating a new recipe.
    else:
        if password == password_repeat:
            recipe.passhash = util.make_md5(password)
            flask.session['passhash'] = recipe.passhash
        else:
            raise werkzeug.exceptions.BadRequest("Passwords don't match")
        recipe_id = dao.make_recipe_id()
        recipe.recipe_id = recipe_id
        dao.recipes.create(recipe.toDict()) # FIXME move save functionality to Recipe class
        # FIXME other auth information is in __init__.py
        flask.session['passhash'] = recipe.passhash

    # Clear the entire HXL Proxy cache to avoid stale data (!!!)
    # TODO: be more targeted here
    cache.clear()

    # Redirect to the /data view page
    return flask.redirect(util.data_url_for(destination_facet, recipe), 303)

# has tests
@app.route("/actions/validate", methods=['POST'])
@util.structlogged
def do_data_validate():
    """ Flask controller: validate an uploaded file against an uploaded HXL schema
    This controller was created for HDX Data Check, which is the only known user.
    The controller returns a JSON validation report from libhxl-python.

    Post parameters:

    url - the URL of the data to validate (required unless "content" is specified)
    content - a file attachment with the HXL content (required unless "url" is specified)
    sheet_index - the 0-based index of the tab in a dataset Excel sheet
    selector - the top-level key for a JSON dataset

    schema_url - the URL of the HXL schema to use (optional; exclusive with "schema_content")
    schema_content - a file attachment with the HXL schema to use (optional; exclusive with "schema_url")
    schema_sheet_index - the 0-based index of the  tab in a schema Excel sheet

    include_dataset - if specified, include the original dataset in the JSON validation result
    """
    flask.g.output_format = 'json' # for error reporting

    # dataset-related POST parameters
    url = flask.request.form.get('url')
    content = flask.request.files.get('content')
    content_hash = None
    if content is not None:
        # need a hash of the content for caching
        content_hash = util.make_file_hash(content)
    sheet_index = flask.request.form.get('sheet', None)
    if sheet_index is not None:
        try:
            sheet_index = int(sheet_index)
        except:
            logup('Bad sheet index', {"sheet": flask.request.form.get('sheet')}, level='warning')
            logger.warning("Bad sheet index: %s", flask.request.form.get('sheet'))
            sheet_index = None
    selector = flask.request.form.get('selector', None)

    # schema-related POST parameters
    schema_url = flask.request.form.get('schema_url')
    schema_content = flask.request.files.get('schema_content')
    schema_content_hash = None
    if schema_content is not None:
        # need a hash of the schema content for caching
        schema_content_hash = util.make_file_hash(schema_content)
    schema_sheet_index = flask.request.form.get('schema_sheet', None)
    if schema_sheet_index is not None:
        try:
            schema_sheet_index = int(schema_sheet_index)
        except:
            logup('Bad schema_sheet index', {"sheet": flask.request.form.get('schema_sheet')}, level='warning')
            logger.warning("Bad schema_sheet index: %s", flask.request.form.get('schema_sheet'))
            schema_sheet_index = None

    # general POST parameters
    include_dataset = flask.request.form.get('include_dataset', False)

    # run the validation and save a report
    # caching happens in the util.run_validation() function
    report = validate.run_validation(
        url, content, content_hash, sheet_index, selector,
        schema_url, schema_content, schema_content_hash, schema_sheet_index,
        include_dataset, flask.request.form
    )

    # return a JSON version of the validation report as an HTTP response
    # (make sure we have a CORS header)
    response = flask.Response(
        json.dumps(
            report,
            indent=4
        ),
        mimetype='application/json'
    )

    # add the CORS header for cross-origin compatibility
    response.headers['Access-Control-Allow-Origin'] = '*'

    # render the JSON response
    return response


# needs tests
# NOTE: This is an experiment that's probably not used anywhere right now
# We may choose to remove it
@app.route('/actions/json-spec', methods=['POST'])
@util.structlogged
def do_json_recipe():
    """ POST handler to execute a JSON recipe
    This POST endpoint allows the user to upload a JSON HXL recipe
    and execute it. The endpoint does NOT currently allow uploading
    a file (but we should add that, to support private datasets).

    POST parameters:

    recipe - a file upload containing a JSON recipe
    format - "csv" or "json"
    show_headers - if specified, include text headers in the output
    use_objects - if specified, use JSON list of objects format
    stub - root filename for downloads

    Information on JSON recipes is available at
    https://github.com/HXLStandard/hxl-proxy/wiki/JSON-recipes
    """

    # Get the JSON recipe as a file attachment
    json_recipe_file = flask.request.files.get('recipe', None)
    if json_recipe_file is None:
        raise werkzeug.exceptions.BadRequest("Parameter 'recipe' is required")
    json_recipe = json.load(json_recipe_file.stream)

    # Other parameters
    format = flask.request.form.get('format', 'csv')
    show_headers = False if flask.request.form.get('show_headers', None) is None else True
    use_objects = False if flask.request.form.get('use_objects', None) is None else True
    stub = flask.request.form.get('stub', 'data')

    # Set global output format
    flask.g.output_format = 'format'

    # Create a HXL filter chain by parsing the JSON recipe
    source = hxl.input.from_spec(json_recipe)

    # Create a JSON or CSV response object, as requested
    if format == 'json':
        response = flask.Response(
            source.gen_json(show_headers=show_headers, use_objects=use_objects),
            mimetype='application/json'
        )
    else:
        response = flask.Response(
            source.gen_csv(show_headers=show_headers),
            mimetype='text/csv'
        )

    # Add the CORS header for cross-origin compatibility
    response.headers['Access-Control-Allow-Origin'] = '*'

    # If a stub is specified, use it to define a download filename
    if stub:
        response.headers['Content-Disposition'] = 'attachment; filename={}.{}'.format(stub, format)

    # Render the output
    return response



########################################################################
# Humanitarian.ID controllers
# (not in active use as of 2019-04)
########################################################################

@app.route('/login')
@util.structlogged
def hid_login():
    """ Flask controller: display login page for Humanitarian.ID
    This is distinct from the /data/login page, which accepts a password
    for a DATASET rather than a USER.

    In the future, we can associate saved datasets with users, so they can
    manage them without individual dataset passwords.

    GET parameters:

    from - the URL of the page from which we were redirected
    """
    # set the from path in a session cookie for later use
    flask.session['login_redirect'] = flask.request.args.get('from', '/')

    # redirect to Humanitarian.ID for the actual login form
    return flask.redirect(auth.get_hid_login_url(), 303)


@app.route('/logout')
@util.structlogged
def hid_logout():
    """ Flask controller: kill the user's login session with Humanitarian.ID

    GET parameters:

    from - the URL of the page from which we were redirected
    """
    path = flask.request.args.get('from', '/') # original path where user choose to log out
    flask.session.clear() # clear the login cookie
    flask.flash("Disconnected from your Humanitarian.ID account (browsing anonymously).")
    return flask.redirect(path, 303)


# not currently in use (until we reactivate H.ID support)
@app.route('/settings/user')
@util.structlogged
def user_settings():
    """ Flask controller: show the user's settings from Humanitarian.ID
    """
    if flask.g.member:
        return flask.render_template('settings-user.html', member=flask.g.member)
    else:
        # redirect back to the settings page after login
        # ('from' is reserved, so we need a bit of a workaround)
        args = { 'from': util.data_url_for('user_settings') }
        return flask.redirect(url_for('login', **args), 303)


@app.route('/oauth/authorized2/1')
@util.structlogged
def do_hid_authorisation():
    """Flask controller: accept an OAuth2 token after successful login via Humanitarian.ID

    GET parameters:

    code - the OAuth2 token
    state - the state that we originally passed to Humanitarian.ID (for verification)
    """
    # grab the token
    code = flask.request.args.get('code')

    # grab the state and check if it's the same as the one we saved in a session cookie
    state = flask.request.args.get('state')
    if state != flask.session.get('state'):
        raise Exception("Security violation: inconsistent state returned from humanitarian.id login request")
    else:
        # if OK, clear the session cookie
        flask.session['state'] = None

    # Look up extra info from Humanitarian.ID
    user_info = auth.get_hid_user(code) # look up user info from Humanitarian.ID
    flask.session['member_info'] = user_info
    flask.flash("Connected to your Humanitarian.ID account as {}".format(user_info.get('name')))

    # Try to bring the user back where s/he started.
    redirect_path = flask.session.get('login_redirect', '/')
    del flask.session['login_redirect']
    return flask.redirect(redirect_path, 303)



########################################################################
# /admin controllers
########################################################################


# needs tests
@app.route("/admin/login")
@util.structlogged
def admin_login():
    """ Log in to use admin functions """
    return flask.render_template('admin-login.html')


# needs tests
@app.route("/admin/recipes/<recipe_id>/")
@util.structlogged
def admin_recipe_view(recipe_id):
    """ View a specific recipe """
    admin.admin_auth()
    recipe = recipes.Recipe(recipe_id, auth=False)
    if 'authorization_token' in recipe.args:
        clone_url = None
    else:
        clone_url = util.data_url_for('data_view', recipe, cloned=True)
    return flask.render_template('admin-recipe-view.html', recipe=recipe, clone_url=clone_url)


# needs tests
@app.route("/admin/recipes/<recipe_id>/edit.html")
@util.structlogged
def admin_recipe_edit(recipe_id):
    """ Edit a saved recipe """
    admin.admin_auth()
    recipe = recipes.Recipe(recipe_id, auth=False)
    args = json.dumps(recipe.args, indent=4)
    return flask.render_template('admin-recipe-edit.html', recipe=recipe, args=args)


# needs tests
@app.route("/admin/recipes/<recipe_id>/delete.html")
@util.structlogged
def admin_recipe_delete(recipe_id):
    """ Delete a saved recipe """
    admin.admin_auth()
    recipe = recipes.Recipe(recipe_id, auth=False)
    return flask.render_template('admin-recipe-delete.html', recipe=recipe)


# needs tests
@app.route("/admin/recipes/")
@util.structlogged
def admin_recipe_list():
    """ List all saved recipes """
    admin.admin_auth()
    recipes = admin.admin_get_recipes()
    return flask.render_template('admin-recipe-list.html', recipes=recipes)


# needs tests
@app.route("/admin/")
@util.structlogged
def admin_root():
    """ Root of admin pages """
    admin.admin_auth()
    return flask.render_template('admin-root.html')


# needs tests
@app.route("/admin/actions/login", methods=['POST'])
@util.structlogged
def do_admin_login():
    """ POST controller for an admin login """
    password = flask.request.form.get('password')
    admin.do_admin_login(password)
    flask.flash("Logged in as admin")
    return flask.redirect('/admin/', 303)


# needs tests
@app.route("/admin/actions/logout", methods=['POST'])
@util.structlogged
def do_admin_logout():
    """ POST controller for an admin logout """
    admin.admin_auth()
    admin.do_admin_logout()
    flask.flash("Logged out of admin functions")
    return flask.redirect('/data/source', 303)


# needs tests
@app.route("/admin/actions/update-recipe", methods=['POST'])
@util.structlogged
def do_admin_update_recipe():
    admin.admin_auth()
    recipe_id = flask.request.form.get('recipe_id')
    admin.do_admin_update_recipe(dict(flask.request.form))
    flask.flash("Updated recipe {}".format(recipe_id))
    return flask.redirect('/admin/recipes/{}/'.format(recipe_id), 303)


# needs tests
@app.route("/admin/actions/delete-recipe", methods=['POST'])
@util.structlogged
def do_admin_delete_recipe():
    admin.admin_auth()
    recipe_id = flask.request.form.get('recipe_id')
    admin.do_admin_delete_recipe(recipe_id)
    flask.flash("Deleted recipe {}".format(recipe_id))
    return flask.redirect('/admin/recipes/'.format(recipe_id), 303)



########################################################################
# Controllers for extra API calls
#
# Migrating to /api (gradually)
#
# None of this is core to the Proxy's function, but this is a convenient
# place to keep it.
########################################################################

@app.route("/api/from-spec.<format>")
@util.structlogged
def from_spec(format="json"):
    """ Use a JSON HXL spec
    Not cached
    """

    # allow format override
    if format != "html":
        format = flask.request.args.get("format", format)
        flask.g.output_format = format

    # other args
    http_headers = {
        'User-Agent': 'hxl-proxy/download'
    }
    filename = flask.request.args.get('filename')
    force = flask.request.args.get("force")

    # check arg logic
    spec_url = flask.request.args.get("spec-url")
    spec_json = flask.request.args.get("spec-json")
    spec = None

    if format == "html":
        return flask.render_template(
            'api-from-spec.html',
            spec_json=spec_json,
            spec_url=spec_url,
            verify_ssl=util.check_verify_ssl(flask.request.args),
            filename=filename,
            force=force,
        )
    elif spec_url and spec_json:
        raise ValueError("Must specify only one of &spec-url or &spec-json")
    elif spec_url:
        spec_response = requests.get(spec_url, verify=verify_ssl, headers=http_headers)
        spec_response.raise_for_status()
        spec = spec_response.json()
    elif spec_json:
        spec = json.loads(spec_json)
    else:
        raise ValueError("Either &spec-url or &spec-json required")

    # process the JSON spec
    source = hxl.input.from_spec(spec)

    # produce appropriate output
    if format == "json":
        response = flask.Response(
            source.gen_json(
                show_headers=spec.get("show_headers", True),
                show_tags=spec.get("show_tags", True),
                use_objects=False
            ),
            mimetype="application/json"
        )
    elif format == "objects.json":
        response = flask.Response(
            source.gen_json(
                show_headers=spec.get("show_headers", True),
                show_tags=spec.get("show_tags", True),
                use_objects=True
            ),
            mimetype="application/json"
        )
    elif format == "csv":
        response = flask.Response(
            source.gen_csv(
                show_headers=spec.get("show_headers", True),
                show_tags=spec.get("show_tags", True)
            ),
            mimetype="text/csv"
        )

    else:
        raise ValueError("Unsupported output format {}".format(format))

    # Add CORS header and return
    response.headers['Access-Control-Allow-Origin'] = '*'

    # Set the file download name if &filename is present
    if filename:
        response.headers['Content-Disposition'] = 'attachment; filename={}'.format(filename)

    return response



# needs tests
@app.route("/api/hxl-test.<format>")
@app.route("/api/hxl-test")
@app.route("/hxl-test.<format>") # legacy path
@app.route("/hxl-test") # legacy path
@util.structlogged
def hxl_test(format='html'):
    """ Flask controller: test if a resource is HXL hashtagged
    GET parameters:
    url - the URL of the resource th check
    @param format: the format for rendering the result.
    """
    flask.g.output_format = format # save the data format for error reporting

    # get the URL
    url = flask.request.args.get('url')
    if not url and (format != 'html'):
        # if it's a web page, show a form; otherwise, throw an error
        raise ValueError("&url parameter required")

    # start the result status report
    result = {
        'status': False,
        'url': url
    }

    # internal function: serialise an exception for inclusion in the report
    def record_exception(e):
        result['exception'] = e.__class__.__name__
        result['args'] = [str(arg) for arg in e.args]

    # if there's a URL, test the resource
    if url:
        try:
            # we grab the columns to force lazy parsing
            util.hxl_data(url, util.make_input_options(flask.request.args)).columns
            # if we get to here, it's OK
            result['status'] = True
            result['message'] = 'Dataset has HXL hashtags'
        except IOError as e1:
            # can't open resource to check it
            result['message'] = 'Cannot load dataset'
            record_exception(e1)
        except hxl.input.HXLTagsNotFoundException as e2:
            # not hashtagged
            result['message'] = 'Dataset does not have HXL hashtags'
            record_exception(e2)
        except BaseException as e3:
            # something else went wrong
            result['message'] = 'Undefined error'
            record_exception(e3)
    else:
        # no URL, so no result
        result = None

    if format == 'json':
        # render a JSON result
        return flask.Response(json.dumps(result), mimetype='application/json')
    else:
        # render an HTML page
        return flask.render_template('hxl-test.html', result=result)


# has tests
@app.route('/api/data-preview.<format>')
#@cache.cached(key_prefix=util.make_cache_key, unless=util.skip_cache_p) # can't cache generator output
@util.structlogged
def data_preview (format="json"):
    """ Return a raw-data preview of any data source supported by the HXL Proxy
    Does not attempt HXL processing.
    """

    def json_generator ():
        """ Generate JSON output, row by row """
        counter = 0
        yield '['
        for row in input:
            if rows > 0 and counter >= rows:
                break
            if counter == 0:
                line = "\n  "
            else:
                line = ",\n  "
            counter += 1
            line += json.dumps(row)
            yield line
        yield "\n]"

    def json_object_generator ():
        """ Generate JSON object-style output, row by row """
        counter = 0
        headers = None
        yield '['
        for row in input:
            if headers is None:
                headers = row
                continue
            if rows > 0 and counter >= rows:
                break
            if counter == 0:
                line = "\n  "
            else:
                line = ",\n  "
            counter += 1
            object = {}
            for i, header in enumerate(headers):
                if header and i < len(row):
                    object[header] = row[i]
            line += json.dumps(object)
            yield line
        yield "\n]"

    def csv_generator ():
        """ Generate CSV output, row by row """
        counter = 0
        for row in input:
            if rows > 0 and counter >= rows:
                break
            counter += 1
            output = io.StringIO()
            csv.writer(output).writerow(row)
            s = output.getvalue()
            output.close()
            yield s

    # allow overriding the format in a parameter (useful for forms)
    if "format" in flask.request.args and format != "html":
        format = flask.request.args.get("format")

    flask.g.output_format = format # for error reporting

    # params
    url = flask.request.args.get('url')

    rows = flask.request.args.get('rows', 0)
    rows = int(rows)

    filename = flask.request.args.get('filename')

    if format == "html":
        return flask.render_template('api-data-preview.html', url=url, args=flask.request.args)

    # if there's no URL, then show an interactive form
    if not url:
        return flask.redirect('/api/data-preview.html', 302)

    # make input
    if util.skip_cache_p():
        input = util.hxl_make_input(url, util.make_input_options(flask.request.args))
    else:
        with caching.input():
            input = util.hxl_make_input(url, util.make_input_options(flask.request.args))

    # Generate result
    if format == 'json':
        response = flask.Response(json_generator(), mimetype='application/json')
    elif format == 'objects.json':
        response = flask.Response(json_object_generator(), mimetype='application/json')
    elif format == 'csv':
        response = flask.Response(csv_generator(), mimetype='text/csv')
    else:
        raise ValueError("Unsupported &format {}".format(format))

    # Add CORS header and return
    response.headers['Access-Control-Allow-Origin'] = '*'

    # Set the file download name if &filename is present
    if filename:
        response.headers['Content-Disposition'] = 'attachment; filename={}'.format(filename)

    return response

# has no tests
@app.route('/api/data-preview-sheets.<format>')
# @cache.cached(key_prefix=util.make_cache_key, unless=util.skip_cache_p) # can't cache generator output
@util.structlogged
def data_preview_sheets(format="json"):
    """ Return names only for the sheets in an Excel workbook.
    In case of csv it returns one sheet name 'Default'
    You must use data_preview to get the actual sheet contents.
    """

    def json_generator():
        """ Generate JSON output, row by row """
        counter = 0
        yield '['
        for row in input:
            if rows > 0 and counter >= rows:
                break
            if counter == 0:
                line = "\n  "
            else:
                line = ",\n  "
            counter += 1
            line += json.dumps(row)
            yield line
        yield "\n]"

    def csv_generator():
        """ Generate CSV output, row by row """
        counter = 0
        for row in input:
            if rows > 0 and counter >= rows:
                break
            counter += 1
            output = io.StringIO()
            csv.writer(output).writerow([row])
            s = output.getvalue()
            output.close()
            yield s

    flask.g.output_format = format  # for error reporting

    # params
    url = flask.request.args.get('url')
    if not url:
        raise ValueError("&url parameter required")

    rows = -1

    # make input
    _output = []

    args = dict(flask.request.args)

    try:
        for sheet in range(0, SHEET_MAX_NO):
            args['sheet'] = sheet
            if util.skip_cache_p():
                input = util.hxl_make_input(url, util.make_input_options(args))
            else:
                with caching.input():
                    input = util.hxl_make_input(url, util.make_input_options(args))
            if isinstance(input, hxl.input.CSVInput):
                _output.append("Default")
                break
            else:
                if input._sheet and input._sheet.name:
                    _output.append(input._sheet.name)
                else:
                    _output.append(str(sheet))

    except HXLIOException as ex:
        logup("Found the last sheet of the Excel file", level='debug')
        logger.debug("Found the last sheet of the Excel file")

    # Generate result
    input = _output
    if format == 'json':
        response = flask.Response(json_generator(), mimetype='application/json')
    elif format == 'csv':
        response = flask.Response(csv_generator(), mimetype='text/csv')
    else:
        raise ValueError("Unsupported &format {}".format(format))

    # Add CORS header and return
    response.headers['Access-Control-Allow-Origin'] = '*'
    return response


# has tests
@app.route('/api/pcodes/<country>-<level>.csv')
@app.route('/pcodes/<country>-<level>.csv') # legacy path
@cache.cached(timeout=604800) # 1 week cache
@util.structlogged
def pcodes_get(country, level):
    """ Flask controller: look up a list of P-codes from iTOS
    @param country: the ISO3 country code
    @param level: the admin level (e.g. "adm2")
    """
    flask.g.output_format = 'csv' # for error reporting

    # Get the P-codes
    with io.StringIO() as buffer:
        pcodes.extract_pcodes(country, level, buffer)
        response = flask.Response(buffer.getvalue(), mimetype='text/csv')

    # Add a CORS header for cross-origin support
    response.headers['Access-Control-Allow-Origin'] = '*'

    # Render the result
    return response


# has tests
@app.route('/api/hash')
@app.route('/hash') # legacy path
@util.structlogged
def make_hash():
    """ Flask controller: hash a HXL dataset
    GET parameters:
    url - the URL of the dataset to check
    headers_only - if specified, hash only the headers, not the content
    """
    flask.g.output_format = 'json' # for error reporting

    # Get the URL; if not supplied, open an HTML form
    url = flask.request.args.get('url')
    if not url:
        return flask.render_template('hash.html')

    # Check if we're hashing only headers
    headers_only = flask.request.args.get('headers_only')

    # Open the HXL dataset
    source = util.hxl_data(url, util.make_input_options(flask.request.args))

    # Generate the report
    report = {
        'hash': source.columns_hash if headers_only else source.data_hash,
        'url': url,
        'date': datetime.datetime.utcnow().isoformat(),
        'headers_only': True if headers_only else False,
        'headers': source.headers,
        'hashtags': source.display_tags
    }

    # Render the JSON response
    return flask.Response(
        json.dumps(report, indent=4),
        mimetype="application/json"
    )


# has tests
@app.route('/api/source-info')
@util.structlogged
def make_info():
    """ Flask controller: get info for an Excel dataset
    GET parameters:
    url - the URL of the dataset to check
    """
    flask.g.output_format = 'json' # for error reporting

    url = flask.request.args.get('url')
    if not url:
        raise ValueError("Parameter 'url' is required")

    # Open the dataset (not necessarily hxlated)
    try:
        info = util.hxl_make_input(url, util.make_input_options(flask.request.args)).info()
        return flask.Response(
            json.dumps(info, indent=4),
            mimetype="application/json"
        )
    except NotImplementedError:
        return flask.Response(
            { "error": "dataset format not supported" },
            mimetype="application/json",
            status=400
        )


########################################################################
# Controllers for removed features (display error messages)
########################################################################


# needs tests
@app.route("/")
@util.structlogged
def home():
    """ Flask controller: nothing currently at root
    Redirect to the /data/source page
    """
    # home isn't moved permanently
    return flask.redirect(flask.url_for("data_source", **flask.request.args) , 302)


# has tests
@app.route('/data/<recipe_id>/chart')
@app.route('/data/chart')
@util.structlogged
def data_chart(recipe_id=None):
    """ Flask controller: discontinued charting endpoint """
    return "The HXL Proxy no longer supports basic charts. Please visit <a href='https://tools.humdata.org/'>tools.humdata.org</a>", 410


# has tests
@app.route('/data/<recipe_id>/map')
@app.route('/data/map')
@util.structlogged
def data_map(recipe_id=None):
    """ Flask controller: discontinued mapping endpoint """
    return "The HXL Proxy no longer supports basic maps. Please visit <a href='https://tools.humdata.org/'>tools.humdata.org</a>", 410


# end



""" Admin functions """

from hxl_proxy import app
from flask import flash, session
from hxl_proxy import dao, exceptions, util

import hashlib, json



########################################################################
# Authentication and authorisation
########################################################################

def admin_auth ():
    """ Check authorisation for admin tasks
    Will redirect to a login page if not authorised.
    ADMIN_PASSWORD_MD5 must be set in the config file
    """
    if not session.get('is_admin'):
        flash("Password required for admin functions")
        raise exceptions.RedirectException('/admin/login')


def do_admin_login (password):
    """ POST action: log into admin functions 
    @param password: the text of the admin password
    """
    expected_passhash = app.config.get('ADMIN_PASSWORD_MD5')
    actual_passhash = hashlib.md5(password.encode("utf-8")).hexdigest()
    if expected_passhash == actual_passhash:
        session['is_admin'] = True
    else:
        flash("Wrong password")
        raise exceptions.Redirect('/admin/login')


def do_admin_logout ():
    """ POST action: log out of admin functions """
    admin_auth()
    session['is_admin'] = False



########################################################################
# Admin functions for recipes
########################################################################

def admin_get_recipes ():
    """ Get a list of saved recipes for the admin page
    @returns: a list of recipes (just dicts, not Recipe objects)
    """
    admin_auth()
    recipes = dao.recipes.list()
    return recipes


def admin_get_recipe (recipe_id):
    """ Look up and return a single recipe for the admin page
    @param recipe_id: the hash id for the saved recipe
    @returns: a Recipe object
    """
    admin_auth()
    recipe = dao.recipes.read(recipe_id)
    return recipe


def do_admin_update_recipe (fields):
    """ POST action: force-update a saved recipe """

    admin_auth()

    # original data fields
    data_fields = dao.recipes.read(fields['recipe_id'])

    # try parsing the JSON args string
    try:
        fields['args'] = json.loads(fields['args'])
    except:
        flash("Parsing error in JSON arguments; restoring old values")
        raise exceptions.RedirectException("/admin/recipes/{}/edit.html".format(fields['recipe_id']))

    # munge the checkbox value
    if fields.get('cloneable') == 'on':
        if 'authorization_token' in fields['args']:
            flash("Cannot make recipe cloneable (contains authorisation token)")
            fields['cloneable'] = False
        else:
            fields['cloneable'] = True
    else:
        fields['cloneable'] = False

    # see if there's a new password
    if fields.get('password'):
        flash("Updated recipe password")
        fields['passhash'] = util.make_md5(fields['password'])
        del fields['password']

    # copy over the new fields
    for key in fields:
        data_fields[key] = fields[key]

    # save to the database
    dao.recipes.update(data_fields)

    
def do_admin_delete_recipe (recipe_id):
    admin_auth()
    dao.recipes.delete(recipe_id)
    



"""Database access functions.

This module contains all database dependencies for the HXL Proxy. It
uses three classes as submodules: L{db} for low-level access, L{user}
for managing user records, and L{recipe} for managing recipe
records. L{user} and L{recipe} both have the standard CRUD (create,
read, update, and delete) functions.  Example:

  from hxl_proxy import dao

  user = dao.users.read('user@example.org')
  user['name_given'] = 'Fred'
  dao.users.update(user)

Database access usually has to happen inside a Flask request context,
and the database connection exists only for the scope of the request;
however, there is a kludge to keep the connection as a local variable
outside a request context for unit testing.

The name of the database is in the Flask config option C{DB_FILE}. The
SQL schema is in the hxl_proxy module directory as C{schema.sql} (see
that file for the properties of each record type).

"""

import sqlite3, mysql.connector, json, os, random, time, base64, hashlib, flask
import hxl_proxy


class db:
    """Low-level database functions"""

    schema_file = hxl_proxy.app.config.get('DB_SCHEMA_FILE', 'schema-sqlite3.sql')

    TEST_SCHEMA_FILE = os.path.join(os.path.dirname(__file__), schema_file)
    """The filename of the SQL schema."""

    type = hxl_proxy.app.config.get('DB_TYPE', 'sqlite3')


    _database = None
    """Internal connection, for testing use outside a request context."""


    @staticmethod
    def connect():
        """Get a database connection

        Will reuse the same connection throughout a request
        context. Uses the C{DB_FILE} Flask config option for the
        location of SQLite3 file.

        @return: a SQLite3 database connection
        """
        if flask.has_request_context(): #FIXME - this is an ugly dependency
            database = getattr(flask.g, '_database', None)
        else:
            database = db._database
        if database is None:
            if db.type == 'sqlite3':
                file = hxl_proxy.app.config.get('DB_FILE', ':memory:')
                database = sqlite3.dbapi2.connect(file)
                database.row_factory = sqlite3.Row
            elif db.type == 'mysql':
                database = mysql.connector.connect(
                    host=hxl_proxy.app.config.get('DB_HOST', 'localhost'),
                    port=hxl_proxy.app.config.get('DB_PORT', 3306),
                    database=hxl_proxy.app.config.get('DB_DATABASE', 'hxl_proxy'),
                    user=hxl_proxy.app.config.get('DB_USERNAME'),
                    password=hxl_proxy.app.config.get('DB_PASSWORD'),
                )
            else:
                raise Exception('Unknown database type: {}'.format(db.DB_TYPE))
            if flask.has_request_context():
                flask.g._database = database
            else:
                db._database = database
        return database

    @hxl_proxy.app.teardown_appcontext
    def close(exception):
        """Close the connection at the end of the request."""
        database = getattr(flask.g, '_database', None)
        if database is not None:
            database.close()

    @staticmethod
    def cursor():
        database = db.connect()
        if db.type == 'mysql':
            return database.cursor(dictionary=True)
        else:
            return database.cursor()

    @staticmethod
    def commit():
        db.connect().commit()

    @staticmethod
    def execute_statement(statement, params=(), commit=False):
        """Execute a single SQL statement, and optionally commit.

        @param statement: the SQL statement to execute.
        @param params: sequence of values for any C{%s} placeholders in the statement.
        @param commit: if True, autocommit at the end of the statement (default: False)
        @return: a SQLite3 cursor object.
        """
        cursor = db.cursor()
        cursor.execute(db.fix_params(statement), params)
        if commit:
            db.commit()
        return cursor

    @staticmethod
    def execute_script(sql_statements, commit=True):
        """Execute a script of statements, and optionally commit.
        @param sql_statements: a string containing multiple SQL statements, separated by ';'
        @param commit: if True, autocommit after executing the statements (default: True)
        @return: a SQLite3 cursor object.
        """
        cursor = db.cursor()
        if db.type == 'mysql':
            for _ in cursor.execute(sql_statements, multi=True):
                pass
        else:
            cursor.executescript(sql_statements)

        if commit:
            db.commit()
        return cursor

    @staticmethod
    def execute_file(filename, commit=True):
        """Open a SQL file and execute it as a script.
        @param filename: path to a file containing SQL statements, separated by ';'
        @param commit: if True, autocommit after executing the statments (default: True)
        @return: a SQLite3 cursor object.
        """
        with open(filename, 'r') as input:
            return db.execute_script(input.read(), commit)

    @staticmethod
    def fetchone(statement, params=()):
        """Fetch a single row of data.
        @param statement: the SQL statement to execute.
        @param params: sequence of values for any placeholders in the statement.
        @return: a single row as a dict.
        @see: L{db.execute_statement}
        """
        row = db.execute_statement(statement, params, commit=False).fetchone()
        if row:
            return dict(row)
        else:
            return None

    @staticmethod
    def fetchall(statement, params=()):
        """Fetch a multiple rows of data.
        @param statement: the SQL statement to execute.
        @param params: sequence of values for any C{%s} placeholders in the statement.
        @return: multiple rows as a list of dicts.
        @see: L{db.execute_statement}
        """
        return [dict(row) for row in db.execute_statement(statement, params, commit=False).fetchall()]

    @staticmethod
    def create_db():
        """Create a new database, erasing the current one.
        Use this method only for temporary databases in unit testing. Uses L{db.TEST_SCHEMA_FILE} to create
        the temporary database.
        """
        db.execute_file(db.TEST_SCHEMA_FILE)

    @staticmethod
    def fix_params(statement):
        if db.type == 'sqlite3':
            statement = statement.replace('%s', '?')
        return statement


class users:
    """Functions for managing database records for users."""

    @staticmethod
    def create(user, commit=True):
        """Add a new user record and optionally commit.
        @param user: a dict of properties for a user.
        @param commit: if True, autocommit after adding the user record (default: True).
        @return: a SQLite3 cursor object.
        """
        return db.execute_statement(
            "insert into Users"
            " (user_id, email, name, name_given, name_family, last_login)"
            " values (%s, %s, %s, %s, %s, %s)",
            (
                user.get('user_id'),
                user.get('email'),
                user.get('name'),
                user.get('name_given'),
                user.get('name_family'),
                time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime()),
            ),
            commit=commit
        )

    @staticmethod
    def read(user_id):
        """Look up a user record by id.
        @param user_id: the user's unique identifier in the database.
        @return: a dict of user properties, or None if the record doesn't exist.
        """
        return db.fetchone(
            'select * from Users where user_id=%s',
            (user_id,)
        )

    @staticmethod
    def update(user, commit=True):
        """Update an existing user record.
        @param user: a dict of user properties, including the C{user_id}.
        @param commit: if True, autocommit after updating the user record (default: True).
        @return: a SQLite3 cursor object.
        """
        return db.execute_statement(
            "update Users"
            " set email=%s, name=%s, name_given=%s, name_family=%s, last_login=%s"
            " where user_id=%s",
            (
                user.get('email'),
                user.get('name'),
                user.get('name_given'),
                user.get('name_family'),
                time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime()),
                user.get('user_id'),
            ),
            commit=commit
        )

    @staticmethod
    def delete(user_id, commit=True):
        """Delete an existing user record and optionally commit.
        @param user_id: the user's unique identifier in the database.
        @param commit: if True, autocommit after deleting the user record (default: True).
        @return: a SQLite3 cursor object.
        """
        return db.execute_statement(
            "delete from Users where user_id=%s",
            (user_id,),
            commit=commit
        )


class recipes:
    """Database recipe records"""

    @staticmethod
    def create(recipe, commit=True):
        """Add a new recipe record and optionally commit.
        @param recipe: a dict of properties for a recipe.
        @param commit: if True, autocommit after adding the recipe record (default: True).
        @return: a SQLite3 cursor object.
        """
        return db.execute_statement(
            "insert into Recipes"
            " (recipe_id, passhash, name, description, cloneable, stub, args, date_created, date_modified)"
            " values (%s, %s, %s, %s, %s, %s, %s, %s, %s)",
            (
                recipe.get('recipe_id'),
                recipe.get('passhash'),
                recipe.get('name'),
                recipe.get('description'),
                recipe.get('cloneable'),
                recipe.get('stub'),
                json.dumps(recipe.get('args', {})),
                time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime()),
                time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime()),
            ),
            commit=commit
        )

    @staticmethod
    def read(recipe_id):
        """Look up a recipe record by id.
        @param recipe_id: the recipe's unique identifier in the database.
        @return: a dict of recipe properties, or None if the record doesn't exist.
        """
        recipe = db.fetchone(
            'select * from Recipes where recipe_id=%s',
            (recipe_id,)
        )
        if recipe:
            recipe['args'] = json.loads(recipe.get('args'))
        return recipe

    @staticmethod
    def update(recipe, commit=True):
        """Update an existing recipe record.
        @param recipe: a dict of recipe properties, including the C{recipe_id}.
        @param commit: if True, autocommit after updating the recipe record (default: True).
        @return: a SQLite3 cursor object.
        """
        return db.execute_statement(
            "update Recipes"
            " set passhash=%s, name=%s, description=%s, cloneable=%s, stub=%s, args=%s, "
            " date_modified=%s"
            " where recipe_id=%s",
            (
                recipe.get('passhash'),
                recipe.get('name'),
                recipe.get('description'),
                recipe.get('cloneable'),
                recipe.get('stub'),
                json.dumps(recipe.get('args', {})),
                time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime()),
                recipe.get('recipe_id'),
            ),
            commit=commit
        )

    @staticmethod
    def delete(recipe_id, commit=True):
        """Delete an existing recipe record and optionally commit.
        @param recipe_id: the recipe's unique identifier in the database.
        @param commit: if True, autocommit after deleting the recipe record (default: True).
        @return: a SQLite3 cursor object.
        """
        return db.execute_statement(
            "delete from Recipes where recipe_id=%s",
            (recipe_id,),
            commit=commit
        )

    @staticmethod
    def list():
        """List all existing recipes, sorted in descending creation-date order"""
        return db.fetchall("select * from Recipes order by date_created desc")


########################################################################
# Support functions
########################################################################

def gen_recipe_id():
    """
    Generate a pseudo-random, 6-character hash for use as a recipe_id.
    """
    salt = str(time.time() * random.random())
    encoded_hash = hxl_proxy.util.make_md5(salt)
    return encoded_hash[:6]

def make_recipe_id():
    """Make a unique recipe_id for a saved recipe."""
    recipe_id = gen_recipe_id()
    while hxl_proxy.dao.recipes.read(recipe_id):
        recipe_id = gen_recipe_id()
    return recipe_id




"""
Remote authentication support for the HXL Proxy.

Started 2015-12 by David Megginson.
"""

import random
import requests
from flask import session
from hxl_proxy import app

def get_hid_login_url ():
    """Construct the URL for logging into Humanitarian ID."""
    state = str(random.randint(1, 100000))
    session['state'] = state
    return '{base_url}/oauth/authorize?response_type=code&client_id={client_id}&scope=profile&redirect_uri={redirect_uri}&state={state}'.format(
        base_url = app.config.get('HID_BASE_URL'),
        client_id = requests.utils.quote(app.config.get('HID_CLIENT_ID')),
        redirect_uri = requests.utils.quote(app.config.get('HID_REDIRECT_URI')),
        state = requests.utils.quote(state)
    )


def get_hid_user (code):
    """Look up a user from Humanitarian.ID based on an authorization token"""

    # Stage 1: get an access token
    params = {
        'client_id': app.config.get('HID_CLIENT_ID'),
        'client_secret': app.config.get('HID_CLIENT_SECRET'),
        'grant_type': 'authorization_code',
        'redirect_uri': app.config.get('HID_REDIRECT_URI'),
        'code': code
    }
    headers = {
        #'Authorization': 'Basic {secret}'.format(secret=app.config.get('HID_CLIENT_SECRET'))
    }
    response = requests.post(
        '{base_url}/oauth/access_token'.format(
            base_url = app.config.get('HID_BASE_URL')
        ), 
        headers=headers,
        data=params
    )

    if response.status_code != 200:
        raise Exception("Failed to get access token: {}".format(response.reason))

    # Stage 2: get the user data

    access_data = response.json()

    response = requests.get(
        '{base_url}/account.json?access_token={access_token}'.format(
            base_url = app.config.get('HID_BASE_URL'),
            access_token = requests.utils.quote(access_data['access_token'])
        )
    )

    return response.json()

# end



"""
Top-level Flask application for HXL Proxy
David Megginson
January 2015

License: Public Domain
Documentation: http://hxlstandard.org

"""

__version__="1.27.2"
"""Module version number
See https://www.python.org/dev/peps/pep-0396/

"""

#
# Continue with other imports
#
import flask, flask_caching, json, os, requests_cache, werkzeug.datastructures
from . import reverse_proxied


#
# Create the main Flask application object
#
app = flask.Flask(__name__)


#
# Handle subpaths
#
app.wsgi_app = reverse_proxied.ReverseProxied(app.wsgi_app)


#
# General config
#
app.config.from_object('hxl_proxy.default_config')
if os.environ.get('HXL_PROXY_CONFIG'):
    app.config.from_envvar('HXL_PROXY_CONFIG')
app.jinja_env.trim_blocks = True
app.jinja_env.lstrip_blocks = True


#
# Set up logging before any other imports, to get correct format in server logs
#
import logging, logging.config, logging.handlers



# Allow all logging at root level, but filter at WSGI level
# That way, other loggers can choose their own level
# of the logging level

class ExcludeFilter(logging.Filter):

    def __init__ (self, excludes):
        self.excludes = excludes

    def filter (self, record):
        return not (record.name in self.excludes)

logging.config.dictConfig(
    {
        'version': 1,
        'formatters': {
            'default': {
                'format': '[%(asctime)s] %(levelname)s in %(name)s: %(message)s',
            },
            # This is where we can implement JSON formatting later
            'plain': {
                'format': '%(message)s',
            },
        },
        'handlers': {
            # default handler for everything but "hxl.REMOTE_ACCESS"
            'wsgi': {
                'class': 'logging.StreamHandler',
                'stream': 'ext://flask.logging.wsgi_errors_stream',
                'formatter': 'default',
                'level': app.config.get('LOGGING_LEVEL', 'ERROR'),
                'filters': ['skip_remote_access'],
            },
            # special handler for "hxl.REMOTE_ACCESS"
            'remote_access': {
                'class': 'logging.handlers.WatchedFileHandler',
                'filename': app.config.get('REMOTE_ACCESS_LOG_FILE', "/var/log/proxy/hxl.log"),
                'formatter': 'plain',
                'level': app.config.get('REMOTE_ACCESS_LOGGING_LEVEL', 'INFO'),
                'filters': ['remote_access'],
            },
        },
        'filters': {
            # Filter to show everything but "hxl.REMOTE_ACCESS"
            'skip_remote_access': {
                '()': ExcludeFilter,
                'excludes': ['hxl.REMOTE_ACCESS'],
            },
            # Filter to show only messages for "hxl.REMOTE_ACCESS"
            'remote_access': {
                '()': logging.Filter,
                'name': 'hxl.REMOTE_ACCESS',
            },
        },
        'root': {
            # Handlers enabled at the start (app will add at least one more)
            'handlers': ['wsgi', 'remote_access',],
            # Default root logging level (overridden by handlers)
            'level': app.config.get('ROOT_LOGGING_LEVEL', 'INFO'),
        },
    }
)

logger = logging.getLogger(__name__)
""" Python logger for this module """

import structlog
logger = structlog.wrap_logger(logging.getLogger('hxl.REMOTE_ACCESS'))

structlog_processors = [
    structlog.contextvars.merge_contextvars,
    structlog.processors.add_log_level,
    structlog.processors.StackInfoRenderer(),
    structlog.dev.set_exc_info,
    structlog.processors.TimeStamper(fmt="iso"),
    structlog.processors.dict_tracebacks,
]

if app.config.get('ENV_TYPE', 'PROD') != 'PROD':
    structlog_processors.append(structlog.dev.ConsoleRenderer())
else:
    structlog_processors.append(structlog.processors.JSONRenderer())

structlog.configure(structlog_processors)

#
# Set up output cache
#
cache = flask_caching.Cache(app, config=app.config.get('CACHE_CONFIG'))
# (Setting up requests cache dynamically in controllers.py)

#
# Needed to register annotations (save until last)
#
import hxl_proxy.controllers

# end



""" Validation support functions
Started May 2018 by David Megginson
License: Public Domain
"""

import hxl_proxy

from hxl_proxy import util

import hxl, logging, werkzeug


logger = logging.getLogger(__name__)
""" Python logger for this module """


@hxl_proxy.cache.memoize(unless=util.skip_cache_p)
def run_validation(url, content, content_hash, sheet_index, selector, schema_url, schema_content, schema_content_hash, schema_sheet_index, include_dataset, args={}):
    """ Do the actual validation run, using the arguments provided.
    Separated from the controller so that we can cache the result easiler.
    The *_hash arguments exist only to assist with caching.
    @returns: a validation report, suitable for returning as JSON.
    """

    # test for opening error conditions
    if (url is not None and content is not None):
        raise werkzeug.exceptions.BadRequest("Both 'url' and 'content' specified")
    if (url is None and content is None):
        raise werkzeug.exceptions.BadRequest("Require one of 'url' or 'content'")
    if (schema_url is not None and schema_content is not None):
        raise werkzeug.exceptions.BadRequest("Both 'schema_url' and 'schema_content' specified")

    # set up the main data
    if content:
        # TODO: stop using libhxl's make_input directly
        source = util.hxl_data(hxl.input.make_input(content, util.make_input_options(args)))
    else:
        source = util.hxl_data(url, util.make_input_options(args))

    # cache if we're including the dataset in the results (we have to run over it twice)
    if include_dataset:
        source = source.cache()

    # set up the schema (if present)
    schema_args = dict(args)
    schema_args['sheet'] = args.get('schema_sheet', args.get('schema_sheet', None))
    schema_args['selector'] = args.get('schema-selector', args.get('schema_selector', None))
    schema_args['timeout'] = args.get('schema-timeout', args.get('schema_timeout', None))
    schema_args['encoding'] = args.get('schema-encoding', args.get('schema_encoding', None))
    schema_args['expand_merged'] = args.get('schema-expand-merged', args.get('schema_expand_merged', None))

    if schema_content:
        schema_source = util.hxl_data(hxl.input.make_input(schema_content, util.make_input_options(schema_args)))
    elif schema_url:
        schema_source = util.hxl_data(schema_url, util.make_input_options(schema_args))
    else:
        schema_source = None

    # Validate the dataset
    report = hxl.validate(source, schema_source)

    # add the URLs if supplied
    if url:
        report['data_url'] = url
    if sheet_index is not None:
        report['data_sheet_index'] = sheet_index
    if schema_url:
        report['schema_url'] = schema_url
    if schema_args.get('sheet') is not None:
        report['schema_sheet_index'] = schema_args.get('sheet')

    # include the original dataset if requested
    if include_dataset:
        content = []
        content.append([util.no_none(column.header) for column in source.columns])
        content.append([util.no_none(column.display_tag) for column in source.columns])
        for row in source:
            content.append([util.no_none(value) for value in row.values])
        report['dataset'] = content

    return report




"""Exception classes.
The proxy uses exceptions sometimes for non-local gotos, such as forcing a redirect.
"""

class RedirectException(Exception):
    """ Exception caught internally for workflow when we need to redirect to a different page
    """

    def __init__(self, target_url, http_code=302, message=None):
        self.target_url = target_url
        self.http_code = http_code
        self.message = message

        
class ForbiddenContentException(Exception):
    """ Exception raised for forbidden content (e.g., URLs or markup in a description)
    """

    def __init__(self, value, reason, field_name=None):
        self.reason = reason
        self.field_name = field_name

    @property
    def human(self):
        return "One of the form fields you submitted had content that is not allowed."

    @property
    def message(self):
        if self.field_name:
            return "Forbidden value for {}: {}".format(self.field_name, self.reason)
        else:
            return "Forbidden value: {}".format(self.reason)

        
class DomainNotAllowedError (IOError):
    """ Error for when a domain is not in the allowed list.
    """

    def __init__ (self, message):
        super().__init__(message)


class RemoteDataException:
    """ 
    Wrapper exception to hide information about a remote data-access failure.
    This prevents a bad actor from using the HXL Proxy to ping a remote
    website (for example).

    """

    def __init__ (self, e):
        self.url = None
        if hasattr(e, 'url'):
            self.url = e.url

    @property
    def human (self):
        return "Sorry, we can't find the remote data you want to process."

    @property
    def message (self):
        if self.url:
            return "No usable data found at {}".format(self.url)
        else:
            return "No usable data found at remote URL"



# end



#!/usr/bin/python
"""Install, build, or test the HXL Proxy.
For details, try
  python setup.py -h
"""

import sys, setuptools

if sys.version_info.major != 3:
    raise SystemExit("The HXL Proxy requires Python 3.x")

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = 'hxl-proxy',
    packages = ['hxl_proxy'],
    package_data={'hxl_proxy': ['*.sql']},
    version = "1.27.2",
    description = 'Flask-based web proxy for HXL',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='David Megginson',
    author_email='contact@megginson.com',
    url='https://github.com/HXLStandard/hxl-proxy',
    include_package_data = True,
    zip_safe = False,
    install_requires=[
        'libhxl==4.27.2',
        #'libhxl @ git+https://github.com/HXLStandard/libhxl-python.git@dev',
        'requests>=2.27',
        'ckanapi>=3.5',
        'flask==2.1.2',
        'flask-caching',
        'requests_cache',
        'mysql-connector-python==8.0.29',
        'redis',
        'urllib3<1.27,>=1.21.1',
        'structlog==22.1.0',
    ],
    dependency_links=[
        "git+https://github.com/HXLStandard/libhxl-python@dev#egg=libhxl",
    ],
    test_suite = "tests",
    tests_require = ['mock']
)



""" Utility functions for hxl_proxy
TODO: also contains template output functions, which need to move to their own module.
Started 2015-02-18 by David Megginson
License: Public Domain
"""

from ast import Try
import hxl_proxy

import flask, hashlib, hxl, json, logging, pickle, random, re, requests, time, urllib
from hxl_proxy import caching, exceptions

from urllib.parse import urlparse

# Logger for this module
logger = logging.getLogger(__name__)

from structlog import contextvars
from functools import wraps
import uuid
from hxl.util import logup

def structlogged(f):
    """ decorator to add essential fields on json logs
    """
    @wraps(f)
    def decorated_structlog_binder(*args, **kwargs):
        contextvars.clear_contextvars()
        contextvars.bind_contextvars(
            user_agent=flask.request.headers.get('User-Agent', "UNKNOWN"),
            peer_ip=flask.request.headers.get('X-Real-IP', flask.request.remote_addr),
            request=flask.request.url,
            request_id=str(uuid.uuid4()),
        )
        return f(*args, **kwargs)
    return decorated_structlog_binder


########################################################################
# Utility functions for caching
########################################################################

def make_cache_key (path = None, args_in=None):
    """ Make a key for a caching request, based on the full path.

    The cache key depends on the path and the GET parameters,
    excluding &force (so that we can cache the request by using
    the force parameter).

    Args:
        path(str): the HTTP path, or None to use the current request
        args_in(dict): the GET params, or None to use the current request

    Returns:
        str: the cache key string

    """
    CACHE_KEY_EXCLUDES = ['force']

    # Fill in default
    if path is None:
        path = flask.request.path
    if args_in is None:
        args_in = flask.request.args

    # Construct the args for the cache key
    args_out = {}
    for name in sorted(args_in.keys()):
        if name not in CACHE_KEY_EXCLUDES:
            args_out[name] = args_in[name]

    # Use a pickled version path and args for the cache key
    return path + pickle.dumps(args_out).decode('latin1')


def skip_cache_p ():
    """ Determine whether we are skipping the cache.

    The HTTP &force parameter requests cache skipping.  Note that this
    function will not detect the &force parameter inside a saved
    recipe; for now, it has to be specified as a GET param.

    Returns:
        bool: True if we don't want to cache

    """
    return True if flask.request.args.get('force') else False



########################################################################
# Input wrappers and options
########################################################################

def hxl_data (raw_source, input_options=None):
    """ Wrapper for hxl.data() to implement a domain-based allow list.

    If the raw_source is a string (i.e. URL), check that the base
    domain is in the allow list. If the allow list is empty, assume
    testing and allow any domain.

    Args:
        raw_source: a HXL data provider, file object, array, or string (representing a URL)
        input_options (hxl.input.InputOptions): input options for reading a dataset

    Returns:
        hxl.model.Dataset: a HXL dataset object

    Raises:
        hxl_proxy.exceptions.DomainNotAllowedError: if the domain for a URL is not in the allow list

    """
    # don't catch exceptions here; see controllers.py for general exception handling
    check_allowed_domain(raw_source)
    return hxl.data(raw_source, input_options)


def hxl_make_input (raw_source, input_options=None):
    """ Wrapper for hxl.make_input() to implement a domain-based allow list.

    If the raw_source is a string (i.e. URL), check that the base
    domain is in the allow list. If the allow list is empty, assume
    testing and allow any domain.

    Args:
        raw_source: a HXL data provider, file object, array, or string (representing a URL)
        input_options(hxl.input.InputOptions): input options for reading a dataset

    Returns:
        hxl.input.AbstractInput: a row-by-row input object (pre-HXL-processing)

    Raises:
        hxl_proxy.exceptions.DomainNotAllowedError: if the domain for a URL is not in the allow list

    """
    # don't catch exceptions here; see controllers.py for general exception handling
    check_allowed_domain(raw_source)
    return hxl.make_input(raw_source, input_options)


def check_allowed_domain (raw_source):
    """ Raise an exception if raw_source is a URL and its base domain is not in the allow list.

    Args:
        raw_source: a HXL data source, possibly a string representing a URL

    Raises:
        hxl_proxy.exceptions.DomainNotAllowedException: if the base domain for a URL is not in the allow list.

    """

    if isinstance(raw_source, str) and not is_allowed_domain(raw_source):
        logup("Domain not in allow list", {'raw_source': raw_source}, "error")
        logger.error("Domain not in allow list: %s", raw_source)
        raise exceptions.DomainNotAllowedError(
            "The HXL Proxy does not allow data from this domain.\n" +
            "If your data is on HDX, you can use the dataset or resource\n" +
            "URL with the HXL Proxy. Otherwise, please send requests to\n"
            "allow new humanitarian-data domains to hdx@un.org.\n\n" +
            raw_source
        )


def is_allowed_domain (raw_source):
    """ Check if raw_source is a URL and its base domain is in the allow list.

    Args:
        raw_source: a HXL data source, possibly a string representing a URL

    Returns:
        bool: True if the url is in the allow list, or the list is empty

    """
    hostnames = hxl_proxy.app.config.get('ALLOWED_DOMAINS_LIST', [])

    # if allowed list is empty just wave in for eveybody. test mode eh?
    # also, if it's not a string (i.e. URL), proceed
    if len(hostnames) == 0 or not isinstance(raw_source, str):
        return True

    url=urlparse(raw_source)
    # it is a hostname match?
    if url.netloc in hostnames:
        return True
    # maybe an allowed domain child?
    domains = tuple(['.{}'.format(d) for d in hostnames])
    if url.netloc.endswith(domains):
        return True
    # sorry, call us
    return False


def make_input_options (args):
    """ Create an InputOptions object from the parameters provided

    Ensure that allow_local is always false. Allow both "-" and "_" between words.

    Args:
        args(dict): an dictionary of command-line params

    Returns:
        hxl.input.InputOptions: the input options extracted from the params

    """

    # set the user agent, to help with analytics
    http_headers = args.get("http-headers", args.get("http_headers", {}))
    http_headers['User-Agent'] = 'hxl-proxy'

    # add an authorisation token, if needed
    if args.get('authorization_token'):
        http_headers['Authorization'] = args['authorization_token']

    # set up the sheet index
    sheet_index = args.get("sheet")
    if sheet_index is not None:
        try:
            sheet_index = int(sheet_index)
        except:
            logup('Invalid sheet index; defaulting to 0', {sheet_index: sheet_index}, level='error')
            logger.error("Invalid sheet index \"%s\"; defaulting to 0", sheet_index)
            sheet_index = 0

    # get the max timeout from configuration, defaulting to 30 seconds
    max_timeout = hxl_proxy.app.config.get("MAX_REQUEST_TIMEOUT", 30.0)
    try:
        timeout = min(float(args.get("timeout")), max_timeout)
    except TypeError:
        timeout = max_timeout

    return hxl.input.InputOptions(
        allow_local = False,
        sheet_index = sheet_index,
        timeout = timeout,
        verify_ssl = check_verify_ssl(args),
        http_headers = http_headers,
        selector = args.get("selector", None),
        encoding = args.get("encoding", None),
        expand_merged = args.get("expand-merged", args.get("expand_merged", False)),
        scan_ckan_resources = args.get("scan-ckan-resources", args.get("scan_ckan_resources", False)),
    )


def check_verify_ssl(args):
    """ Check parameters to see if we need to verify SSL connections.

    Args:
        args(dict): command-line parameters

    Returns:
        bool: True if we need to verify SSL; false if unverified is OK

    """
    if args.get('skip-verify-ssl', args.get("skip_verify_ssl", None)) == 'on':
        return False
    elif args.get('verify_ssl') == 'off' or args.get('verify') == 'off': # deprecated parameters
        return False
    else:
        return True



########################################################################
# Not yet classified
########################################################################

def check_markup(s):
    """ Check if text contains markup or URLs
    """
    if re.search("(<[a-zA-Z]|>|href|https?:|\.com|\[url=)", s):
        return True
    else:
        return False

def forbid_markup(s, field_name=None):
    """ Raise an exception if markup or URLs appear in a string
    Otherwise, return the string.
    """
    if check_markup(s):
        raise hxl_proxy.exceptions.ForbiddenContentException(s, "markup and URLs not allowed", field_name)
    else:
        return s

def clean_tagger_mappings(headers, recipe):
    """ Create a clean list of hashtag mappings for the tagger form
    """
    existing_mappings = dict()
    mappings = list()
    headers_seen = set()

    # Existing mappings are in tagger-nn-header and tagger-nn-tag
    for arg in recipe.args:
        result = re.match(r'^tagger-(\d{2})-header', arg)
        if result:
            n = result.group(1)
            header = hxl.datatypes.normalise_string(recipe.args[arg])
            tagspec = recipe.args.get("tagger-" + n + "-tag")
            if header and tagspec and header not in existing_mappings:
                existing_mappings[header] = tagspec

    # Do the headers in document order
    for header in headers:
        header = hxl.datatypes.normalise_string(header)
        if header and header not in headers_seen:
            mappings.append([header, existing_mappings.get(header, "")])
            headers_seen.add(header)

    # Add any extra mappings back in
    for header in existing_mappings:
        if header not in headers_seen:
            mappings.append([header, existing_mappings[header]])
            headers_seen.add(header)

    return mappings


def make_file_hash(stream):
    """Calculate a hash in chunks from a stream.
    Must be random-access. Resets to position 0 before and after read.
    @param stream: a bytes stream to read
    @returns: an MD5 hash
    """
    file_hash = hashlib.md5()
    stream.seek(0)
    while True:
        s = stream.read(8192)
        if len(s) < 1:
            break
        else:
            file_hash.update(s)
    stream.seek(0)
    return file_hash.hexdigest()

def make_md5(s):
    """Return an MD5 hash for a string."""
    return hashlib.md5(s.encode('utf-8')).hexdigest()

def make_args(recipe=None, format=None, flavour=None, recipe_id=None, cloned=False):
    """Construct args for url_for."""
    args = {}
    if recipe and recipe.args and (cloned or not recipe_id):
        for key in recipe.args:
            args[key] = recipe.args[key]
    if format:
        args['format'] = format
        args['flavour'] = flavour
        stub = recipe.stub
        if stub:
            args['stub'] = re.sub("[^a-zA-Z0-9_-]+", '_', stub) # Flask's url_for doesn't escape !!!
    if not recipe_id and recipe:
        recipe_id = recipe.recipe_id
    if recipe_id and not cloned:
        args['recipe_id'] = recipe_id
    return args

def make_json_error(e, status):
    """Convert an exception to a string containing a JSON-formatted error
    @param e: the exception to convert
    @returns: a string containing JSON markup
    """
    json_error = {
        'error': e.__class__.__name__,
        'status': status
    }
    if hasattr(e, 'message'):
        json_error['message'] = e.message
    else:
        json_error['message'] = str(e)
    if hasattr(e, 'human'):
        json_error['human_message'] = e.human
    if hasattr(e, 'errno') and (e.errno is not None):
        json_error['errno'] = e.errno
    if hasattr(e, 'response') and e.response:
        json_error['source_status_code'] = e.response.status_code
        json_error['source_url'] = e.response.url
        json_error['source_message'] = e.response.text
    return json.dumps(json_error, indent=4, sort_keys=True)


########################################################################
# Output helper functions for templates
########################################################################

def no_none(s):
    """ Template helper: prevent None from appearing in output.
    @param s: the input value
    @returns: a string version of the value, or "" if None
    """
    return str(s) if s is not None else ''

def stream_template(template_name, **context):
    """From the flask docs - stream a long template result."""
    hxl_proxy.app.update_template_context(context)
    t = hxl_proxy.app.jinja_env.get_template(template_name)
    rv = t.stream(context)
    rv.enable_buffering(5)
    return rv

def urlquote(value):
    return urllib.parse.quote_plus(value, safe='/')

def spreadsheet_col_num_to_name(num):
    """Convert a column index to spreadsheet letters.
    Adapted from http://asyoulook.com/computers%20&%20internet/python-convert-spreadsheet-number-to-column-letter/659618
    """
    letters = ''
    num += 1
    while num:
        mod = num % 26
        letters += chr(mod + 64)
        num = num // 26
    return ''.join(reversed(letters))

def using_tagger_p(recipe):
    for name in recipe.args:
        if re.match(r'^tagger-', name):
            return True
    return False

def add_args(extra_args):
    """Add GET parameters."""
    args = {key: flask.request.args.get(key) for key in flask.request.args}
    for key in extra_args:
        if extra_args[key]:
            # add keys with truthy values
            args[key] = extra_args[key]
        else:
            # remove keys with non-truthy values
            if args.get(key):
                del args[key]
    return '?' + urlencode_utf8(args)

def data_url_for(endpoint, recipe=None, format=None, flavour=None, recipe_id=None, cloned=False, extras={}):
    """Generate a URL relative to the subpath (etc)
    Wrapper around flask.url_for
    """
    if not recipe_id and recipe:
        recipe_id = recipe.recipe_id
    args = make_args(recipe, format=format, flavour=flavour, recipe_id=recipe_id, cloned=cloned)
    if extras:
        # add in any extra GET params requested
        for key in extras:
            args[key] = extras[key]
    if recipe_id and not cloned:
        args['recipe_id'] = recipe_id
    return flask.url_for(endpoint, **args)

def severity_class(severity):
    """Return a CSS class for a validation error severity"""
    if severity == 'error':
        return 'severity_error'
    elif severity == 'warning':
        return 'severity_warning'
    else:
        return 'severity_info'

def re_search(regex, string):
    """Try matching a regular expression."""
    return re.search(regex, string)

def search_by_attributes(attributes, columns):
    result_columns = []
    for column in columns:
        for attribute in attributes:
            if attribute in column.attributes:
                result_columns.append(column)
                break
    return result_columns

def get_gravatar(email, size=40):
    import hashlib
    hash = hashlib.md5(email.encode('utf8').lower()).hexdigest()
    url = "http://www.gravatar.com/avatar/{hash}?s={size}".format(
        hash=hash,
        size=size
    )
    return url


def urlencode_utf8(params):
    if hasattr(params, 'items'):
        params = list(params.items())
    return '&'.join(
            urlquote(k) + '=' + urlquote(v) for k, v in params if v
    )

#
# Declare Jinja2 filters and functions
#

hxl_proxy.app.jinja_env.filters['nonone'] = (
    no_none
)

hxl_proxy.app.jinja_env.filters['urlquote'] = (
    urlquote
)

hxl_proxy.app.jinja_env.filters['spreadsheet_col'] = (
    spreadsheet_col_num_to_name
)

hxl_proxy.app.jinja_env.filters['strnorm'] = (
    hxl.datatypes.normalise_string
)

hxl_proxy.app.jinja_env.globals['static'] = (
    lambda filename: flask.url_for('static', filename=filename)
)

hxl_proxy.app.jinja_env.globals['using_tagger_p'] = (
    using_tagger_p
)

hxl_proxy.app.jinja_env.globals['add_args'] = (
    add_args
)

hxl_proxy.app.jinja_env.globals['data_url_for'] = (
    data_url_for
)

hxl_proxy.app.jinja_env.globals['severity_class'] = (
    severity_class
)

hxl_proxy.app.jinja_env.globals['re_search'] = (
    re_search
)

hxl_proxy.app.jinja_env.globals['search_by_attributes'] = (
    search_by_attributes
)

hxl_proxy.app.jinja_env.globals['get_gravatar'] = (
    get_gravatar
)




"""Configure a filter pipeline from command-line arguments.

The main entry point is setup_filters()

The GET parameters have numbers appended, e.g. "rename-oldtag7". This
module uses the numbers to group the parameters, then to construct the
hxl.filter objects from them and build a pipeline.
"""

import hxl, io
from hxl_proxy import exceptions, util
import hxl.filters # why do we have to import this???
from hxl.converters import Tagger


# Maximum number of filters to check
MAX_FILTER_COUNT = 99

def setup_filters(recipe, data_content=None):
    """
    Open a stream to a data source URL, and create a filter pipeline based on the arguments.
    @param: recipe the GET-request recipe (uses only recipe.args).
    @param data_content: a dataset uploaded directly via a POST request.
    @returns: a HXL DataSource representing the full pipeline.
    """

    # null recipe or url means null source
    if not data_content and (not recipe or not recipe.url):
        return None

    # Basic input source

    input_options = util.make_input_options(recipe.args)

    if data_content:
        source = util.hxl_data(io.BytesIO(data_content.encode('utf-8')), input_options)
    else:
        try:
            source = util.hxl_data(recipe.args["url"], input_options)
            source.columns
        except hxl.input.HXLTagsNotFoundException:
            source = util.hxl_data(make_tagged_input(recipe.args, input_options), input_options)

    # Do we have a JSON recipe? Load it first.
    if recipe.args.get('recipe'):
        source = source.recipe(recipe.args.get('recipe'))

    # Intercept missing hashtags here
    try:
        source.columns
    except hxl.input.HXLTagsNotFoundException:
        raise exceptions.RedirectException(util.data_url_for('data_tagger', recipe), 303, 'No HXL hashtags found')

    # Create the filter pipeline from the source
    for index in range(1, MAX_FILTER_COUNT):
        filter = recipe.args.get('filter%02d' % index)
        if filter == 'add':
            source = add_add_filter(source, recipe.args, index)
        elif filter == 'append':
            source = add_append_filter(source, recipe.args, index)
        elif filter == 'append-list':
            source = add_append_list_filter(source, recipe.args, index)
        elif filter == 'clean':
            source = add_clean_filter(source, recipe.args, index)
        elif filter == 'count':
            source = add_count_filter(source, recipe.args, index)
        elif filter == 'column' or filter == 'cut':
            source = add_column_filter(source, recipe.args, index)
        elif filter == 'dedup':
            source = add_dedup_filter(source, recipe.args, index)
        elif filter == 'expand':
            source = add_expand_filter(source, recipe.args, index)
        elif filter == 'explode':
            source = add_explode_filter(source, recipe.args, index)
        elif filter == 'fill':
            source = add_fill_filter(source, recipe.args, index)
        elif filter == 'implode':
            source = add_implode_filter(source, recipe.args, index)
        elif filter == 'jsonpath':
            source = add_jsonpath_filter(source, recipe.args, index)
        elif filter == 'merge':
            source = add_merge_filter(source, recipe.args, index)
        elif filter == 'rename':
            source = add_rename_filter(source, recipe.args, index)
        elif filter == 'replace':
            source = add_replace_filter(source, recipe.args, index)
        elif filter == 'replace-map':
            source = add_replace_map_filter(source, recipe.args, index)
        elif filter == 'rows' or filter == 'select':
            source = add_row_filter(source, recipe.args, index)
        elif filter == 'sort':
            source = add_sort_filter(source, recipe.args, index)
        elif filter:
            raise Exception("Unknown filter type '{}'".format(filter))

    return source

def make_tagged_input(args, input_options):
    """Create the raw input, optionally using the Tagger filter."""
    input = util.hxl_make_input(args.get("url"), input_options)

    # Intercept tagging as a special data input
    specs = []
    for n in range(1, 101):
        header = args.get('tagger-%02d-header' % n)
        tag = _parse_tagspec(args.get('tagger-%02d-tag' % n))
        if header and tag:
            specs.append((header, tag))
    if len(specs) > 0:
        match_all = (True if args.get('tagger-match-all') else False)
        default_tag = _parse_tagspec(args.get('tagger-default-tag'))
        if not default_tag:
            default_tag = None
        input = Tagger(input, specs, default_tag=default_tag, match_all=match_all)

    return input

def add_add_filter(source, args, index):
    """Add the hxladd filter to the end of the chain."""
    tagspec = _parse_tagspec(args.get('add-tag%02d' % index))
    header = args.get('add-header%02d' % index)
    value = args.get('add-value%02d' % index, '')
    before = (args.get('add-before%02d' % index) == 'on')
    values = [(hxl.Column.parse(tagspec, header=header), value)]
    return source.add_columns(specs=values, before=before)

def add_append_filter(source, args, index):
    """Add the hxlappend filter to the end of the chain."""
    exclude_columns = args.get('append-exclude-columns%02d' % index, False)
    append_sources = []
    for subindex in range(1, 100):
        append_source = args.get('append-dataset%02d-%02d' % (index, subindex))
        if append_source:
            append_sources.append(append_source)
    row_query = args.get('append-where%02d' % index, None)
    return source.append(
        append_sources=append_sources,
        add_columns=(not exclude_columns),
        queries=row_query
    )

def add_append_list_filter(source, args, index):
    """Add the hxlappend filter to the end of the chain with an external list."""
    exclude_columns = args.get('append-list-exclude-columns%02d' % index, False)
    source_list_url = args.get('append-list-url%02d' % index, None)
    row_query = args.get('append-list-where%02d' % index, None)
    return source.append_external_list(
        source_list_url=source_list_url,
        add_columns=(not exclude_columns),
        queries=row_query
    )

def add_clean_filter(source, args, index):
    """Add the hxlclean filter to the end of the pipeline."""
    whitespace_tags = hxl.TagPattern.parse_list(args.get('clean-whitespace-tags%02d' % index, ''))
    upper_tags = hxl.TagPattern.parse_list(args.get('clean-toupper-tags%02d' % index, ''))
    lower_tags = hxl.TagPattern.parse_list(args.get('clean-tolower-tags%02d' % index, ''))
    date_tags = hxl.TagPattern.parse_list(args.get('clean-date-tags%02d' % index, ''))
    date_format = args.get('clean-date-format%02d' % index, None);
    number_tags = hxl.TagPattern.parse_list(args.get('clean-num-tags%02d' % index, ''))
    number_format = args.get('clean-number-format%02d' % index, None);
    latlon_tags = hxl.TagPattern.parse_list(args.get('clean-latlon-tags%02d' % index, ''))
    purge_flag = args.get('clean-purge%02d' % index, False)
    row_query = args.get('clean-where%02d' % index, None)
    return source.clean_data(
        whitespace=whitespace_tags,
        upper=upper_tags,
        lower=lower_tags,
        date=date_tags,
        date_format=date_format,
        number=number_tags,
        number_format=number_format,
        latlon=latlon_tags,
        purge=purge_flag,
        queries=row_query
    )

def add_count_filter(source, args, index):
    """Add the hxlcount filter to the end of the pipeline."""
    tags = hxl.TagPattern.parse_list(args.get('count-tags%02d' % index, ''))
    row_query = args.get('count-where%02d' % index, None)

    aggregators = []
    for n in range(1, 25):
        suffix = '%02d-%02d' % (index, n,)
        count_type = args.get('count-type' + suffix)
        if count_type:
            aggregators.append(hxl.filters.Aggregator(
                type = count_type,
                pattern = args.get('count-pattern' + suffix),
                column = hxl.model.Column.parse(
                    _parse_tagspec(args.get('count-column' + suffix, '#meta+' + count_type)),
                    header = args.get('count-header' + suffix, count_type.title())
                )
            ))

    # deprecated parameters

    count_spec = args.get('count-spec%02d' % index, None)
    if count_spec:
        # deprecated column hashtag for a default count column
        aggregators.append(hxl.filters.Aggregator(
            type = 'count',
            column = hxl.model.Column.parse_spec(count_spec, default_header='Count')
        ))

    aggregate_pattern = args.get('count-aggregate-tag%02d' % index)
    if aggregate_pattern:
        if not count_spec:
            aggregators.append(hxl.filters.Aggregator(
                type='count',
                column = hxl.model.Column.parse('#meta+count', header='Count')
            ))
        for count_type in ['sum', 'average', 'min', 'max']:
            aggregators.append(hxl.filters.Aggregator(
                type = count_type,
                pattern = aggregate_pattern,
                column = hxl.model.Column.parse("#meta+" + count_type, header=count_type.title())
            ))

    return source.count(patterns=tags, aggregators=aggregators, queries=row_query)

def add_column_filter(source, args, index):
    """Add the hxlcut filter to the end of the pipeline."""
    include_tags = hxl.TagPattern.parse_list(args.get('cut-include-tags%02d' % index, []))
    exclude_tags = hxl.TagPattern.parse_list(args.get('cut-exclude-tags%02d' % index, []))
    skip_untagged = args.get('cut-skip-untagged%02d' % index, False)
    if include_tags:
        source = source.with_columns(include_tags)
    if exclude_tags or skip_untagged:
        source = source.without_columns(exclude_tags, skip_untagged=skip_untagged)
    return source

def add_dedup_filter(source, args, index):
    tags = args.get('dedup-tags%02d' % index, [])
    row_query = args.get('dedup-where%02d' % index, '')
    return source.dedup(tags, queries=row_query)

def add_expand_filter(source, args, index):
    tags = args.get('expand-tags%02d' % index, [])
    separator = args.get('expand-separator%02d' % index, "|")
    correlate = (args.get('expand-correlate%02d' % index) == 'on')
    row_query = args.get('expand-where%02d' % index, '')
    return source.expand_lists(
        patterns = tags,
        separator=separator,
        correlate=correlate,
        queries=row_query
    )

def add_explode_filter(source, args, index):
    return source.explode(
        args.get('explode-header-att%02d' % index, 'header'),
        args.get('explode-value-att%02d' % index, 'value')
    )

def add_fill_filter(source, args, index):
    patterns = args.get('fill-patterns%02d' % index, None)
    if not patterns:
        # deprecated
        patterns = args.get('fill-pattern%02d' % index, None)
    queries = args.get('fill-where%02d' % index, None)
    return source.fill_data(patterns=patterns, queries=queries)

def add_implode_filter(source, args, index):
    return source.implode(
        label_pattern=args.get('implode-label-pattern%02d' % index, 'header'),
        value_pattern=args.get('implode-value-pattern%02d' % index, 'value')
    )

def add_jsonpath_filter(source, args, index):
    path = args.get('jsonpath-path%02d' % index)
    patterns = args.get('jsonpath-patterns%02d' % index, None)
    queries = args.get('jsonpath-where%02d' % index, None)
    use_json = (args.get('jsonpath-flatten%02d' % index) != 'on')
    return source.jsonpath(path, patterns=patterns, queries=queries, use_json=use_json)

def add_merge_filter(source, args, index):
    """Add the hxlmerge filter to the end of the pipeline."""
    tags = hxl.TagPattern.parse_list(args.get('merge-tags%02d' % index, []))
    keys = hxl.TagPattern.parse_list(args.get('merge-keys%02d' % index, []))
    replace = (args.get('merge-replace%02d' % index) == 'on')
    overwrite = (args.get('merge-overwrite%02d' % index) == 'on')
    url = args.get('merge-url%02d' % index)
    merge_source = util.hxl_data(url, util.make_input_options(args))
    return source.merge_data(merge_source, keys=keys, tags=tags, replace=replace, overwrite=overwrite)

def add_rename_filter(source, args, index):
    """Add the hxlrename filter to the end of the pipeline."""
    oldtag = hxl.TagPattern.parse(args.get('rename-oldtag%02d' % index))
    oldheader = hxl.datatypes.normalise_string(args.get('rename-oldheader%02d' % index))
    tagspec = _parse_tagspec(args.get('rename-newtag%02d' % index))
    header = args.get('rename-header%02d' % index)
    column = hxl.Column.parse(tagspec, header=header)
    return source.rename_columns([(oldtag, column, oldheader)])

def add_replace_filter(source, args, index):
    """Add the hxlreplace filter to the end of the pipeline."""
    original = args.get('replace-pattern%02d' % index)
    replacement = args.get('replace-value%02d' % index)
    tags = args.get('replace-tags%02d' % index)
    use_regex = args.get('replace-regex%02d' % index)
    row_query = args.get('replace-where%02d' % index)
    return source.replace_data(original, replacement, tags, use_regex, queries=row_query)

def add_replace_map_filter(source, args, index):
    """Add the hxlreplace filter to the end of the pipeline."""
    url = args.get('replace-map-url%02d' % index)
    row_query = args.get('replace-map-where%02d' % index)
    return source.replace_data_map(util.hxl_data(url, util.make_input_options(args)), queries=row_query)

def add_row_filter(source, args, index):
    """Add the hxlselect filter to the end of the pipeline."""
    queries = []
    for subindex in range(1, 6):
        query = args.get('select-query%02d-%02d' % (index, subindex))
        if query:
            queries.append(query)
    reverse = (args.get('select-reverse%02d' % index) == 'on')
    if reverse:
        return source.without_rows(queries)
    else:
        return source.with_rows(queries)

def add_sort_filter(source, args, index):
    """Add the hxlsort filter to the end of the pipeline."""
    tags = hxl.TagPattern.parse_list(args.get('sort-tags%02d' % index, ''))
    reverse = (args.get('sort-reverse%02d' % index) == 'on')
    return source.sort(tags, reverse)

def _parse_tagspec(s):
    if not s:
        return None
    elif s[0] == '#':
        return s
    else:
        return '#' + s

# end
